# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Inheritable test scenarios."""

import contextlib
import textwrap
from collections.abc import Generator
from functools import cached_property
from typing import Any, TYPE_CHECKING, override

from django.db.models import Model

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebianAutopkgtest,
    DebianAutopkgtestResult,
    DebianAutopkgtestResultStatus,
    DebianAutopkgtestSource,
    DebianBlhc,
    DebianLintian,
    DebianLintianSummary,
    DebianPiuparts,
    DebianSourcePackage,
    DebianUpload,
    TaskTypes,
)
from debusine.client.models import RuntimeParameter
from debusine.db.context import context
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    Collection,
    CollectionItem,
    Group,
    Scope,
    Token,
    User,
    WorkRequest,
    Worker,
    WorkerPool,
    WorkflowTemplate,
    Workspace,
)
from debusine.db.models.auth import Identity
from debusine.db.models.collections import CollectionRelation
from debusine.db.permissioncontext import PermissionContext
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.models import (
    RegressionTrackingUpdateQAResults,
    SbuildWorkflowData,
)
from debusine.tasks import DebDiff
from debusine.tasks.models import (
    BackendType,
    OutputData,
    RegressionAnalysis,
    RegressionAnalysisStatus,
    SbuildInput,
    WorkerType,
)

if TYPE_CHECKING:
    from debusine.db.playground.playground import Playground


class Scenario:
    """Base for inheritable test scenarios."""

    # Implementation notes:
    #
    # Database is rolled back for each test method after build() is called:
    #
    # * objects created at build() time remain valid in the database (and are
    #   reset to their build() state)
    # * objects created in other methods will become invalid at the end of the
    #   test method
    #
    # This means no @cached_property and other internally cached values that
    # are not created in build().
    #
    # See #626 for details.

    playground: "Playground"

    def __init__(self, *, set_current: bool = False) -> None:
        """Store scenario arguments."""
        self.needs_set_current = set_current

    def build(self, playground: "Playground") -> None:
        """
        Build the scenario.

        This is run by Playground with permission tests disabled and in a local
        context that will be restored at the end of the build.
        """
        self.playground = playground

    def set_current(self) -> None:
        """Set the current user and workspace from the scenario."""


class DefaultScopeUser(Scenario):
    """
    Quick access to the default scope and user.

    Optionally sets them in context.
    """

    scope: Scope
    user: User

    @override
    def build(self, playground: "Playground") -> None:
        super().build(playground)
        self.scope = self.playground.get_default_scope()
        self.user = self.playground.get_default_user()

    @override
    def set_current(self) -> None:
        super().set_current()
        context.set_permission_context_scoped(self.scope, self.context_ui())

    def create_user_token(self) -> Token:
        """Create a user token for self.user."""
        return self.playground.create_user_token(user=self.user)

    @cached_property
    def scope_owners(self) -> Group:
        """Return the group of scope owners."""
        return self.playground.create_group_role(self.scope, Scope.Roles.OWNER)

    @contextlib.contextmanager
    def assign_role(self, resource: Model, *roles: str) -> Generator[None]:
        """Temporarily assign roles to the scenario user on the resource."""
        with self.playground.assign_role(resource, self.user, *roles):
            yield

    def context_ui(self) -> PermissionContext:
        """Return a PermissionContext for this scenario user, in the UI case."""
        return PermissionContext.for_ui_user(self.user)


class DefaultScopeUserAPI(DefaultScopeUser):
    """DefaultScopeUser, plus a user token."""

    user_token: Token

    @override
    def build(self, playground: "Playground") -> None:
        super().build(playground)
        self.user_token = self.create_user_token()


class DefaultContext(DefaultScopeUser):
    """
    Quick access to the default scope, user and workspace.

    Optionally sets them in context.
    """

    workspace: Workspace
    default_task_configuration_collection: Collection

    @override
    def build(self, playground: "Playground") -> None:
        super().build(playground)
        self.workspace = self.playground.get_default_workspace()
        self.default_task_configuration_collection, _ = (
            Collection.objects.get_or_create(
                workspace=self.workspace,
                name="default",
                category=CollectionCategory.TASK_CONFIGURATION,
            )
        )

    @override
    def set_current(self) -> None:
        super().set_current()
        self.workspace.set_current()

    @cached_property
    def workspace_owners(self) -> Group:
        """Return the group of workspace owners."""
        return self.playground.create_group_role(
            self.workspace, Workspace.Roles.OWNER
        )


class DefaultContextAPI(DefaultScopeUserAPI, DefaultContext):
    """DefaultContext, plus a user token."""


class DefaultSbuild(DefaultContext):
    """Scenario with a source and an environment ready for a sbuild task."""

    source: Artifact
    environment_sid: Artifact

    @override
    def build(self, playground: "Playground") -> None:
        super().build(playground)
        self.source = self.playground.create_minimal_source_package_artifact()
        environment = self.playground.create_debian_environment(codename="sid")
        assert environment.artifact is not None
        self.environment_sid = environment.artifact


class DefaultSbuildAPI(DefaultScopeUserAPI, DefaultSbuild):
    """DefaultSbuild, plus a user token."""


class PackageBuildLog(DefaultContext):
    """Workspace with a package build log."""

    @override
    def build(self, playground: "Playground") -> None:
        super().build(playground)

        auotpkgtest_logs_collection, _ = Collection.objects.get_or_create(
            workspace=self.workspace,
            name="autopkgtest-logs",
            category=CollectionCategory.QA_RESULTS,
        )

        # Create an AutopkgtestLog artifact
        self.autopkgtest_log = textwrap.dedent(
            """\
            0s autopkgtest [19:23:39]: starting date and time: 2025-09-17 19:23:39+0000
            0s autopkgtest [19:23:39]: version 5.47~bpo12+1
            0s autopkgtest [19:23:39]: host debusine-worker-amd64-hades-04; command line: /usr/bin/autopkgtest --apt-upgrade --output-dir=artifact-dir --summary=artifact-dir/summary --no-built-binaries --needs-internet=run /tmp/debusine-fetch-exec-upload-6fvpyw4c/phing_3.1.0-1_all.deb /tmp/debusine-fetch-exec-upload-6fvpyw4c/phing_3.1.0-1.dsc -- incus artifact/2482231 -- --profile debusine
            7s autopkgtest [19:23:46]: testbed dpkg architecture: amd64
            7s autopkgtest [19:23:46]: testbed apt version: 3.1.5
            7s autopkgtest [19:23:46]: @@@@@@@@@@@@@@@@@@@@ test bed setup
            7s autopkgtest [19:23:46]: testbed release detected to be: None
            7s autopkgtest [19:23:46]: updating testbed package index (apt update)
            7s Get:1 http://deb.debian.org/debian sid InRelease [216 kB]
            78s autopkgtest [22:28:32]: @@@@@@@@@@@@@@@@@@@@ apt-source debci
            81s Get:1 http://deb.debian.org/debian unstable/main debci 3.12 (dsc) [2949 B]
            81s Get:2 http://deb.debian.org/debian unstable/main debci 3.12 (tar) [164 kB]
            81s dpkg-source: warning: cannot verify inline signature for ./debci_3.12.dsc: no acceptable signature found
            ...
            32s autopkgtest [12:16:27]: test pybuild-autopkgtest: preparing testbed
            32s Ign:1 file:/tmp/autopkgtest.0wEqIN/binaries  InRelease
            32s Get:2 file:/tmp/autopkgtest.0wEqIN/binaries  Release [816 B]
            32s Get:2 file:/tmp/autopkgtest.0wEqIN/binaries  Release [816 B]
            33s Ign:3 file:/tmp/autopkgtest.0wEqIN/binaries  Release.gpg
            33s Reading package lists...
            33s Reading package lists...
            33s Building dependency tree...
            102s autopkgtest [22:28:56]: test pybuild-autopkgtest: [-----------------------
            102s + tmpdir=/tmp/autopkgtest.4eb2Ze/autopkgtest_tmp
            102s + cp -r spec /tmp/autopkgtest.4eb2Ze/autopkgtest_tmp
            102s + cd /tmp/autopkgtest.4eb2Ze/autopkgtest_tmp
            102s + export COVERAGE=no
            102s + rspec
            103s autopkgtest [22:34:24]: test pybuild-autopkgtest:  - - - - - - - - - - results - - - - - - - - - -
            103s pybuild-autopkgtest           PASS
            109s autopkgtest [22:29:03]: test test-suite: preparing testbed
            109s qemu-system-x86_64: terminating on signal 15 from pid 2264680 (/usr/bin/python3)
            129s autopkgtest [22:29:23]: testbed dpkg architecture: amd64
            129s autopkgtest [22:29:23]: testbed apt version: 3.1.6
            130s autopkgtest [22:29:24]: @@@@@@@@@@@@@@@@@@@@ test bed setup
            130s autopkgtest [22:29:24]: updating testbed package index (apt update)
            130s Get:1 http://deb.debian.org/debian unstable InRelease [207 kB]
            130s Get:2 http://deb.debian.org/debian unstable/main Sources.diff/Index [63.6 kB]
            131s Get:3 http://deb.debian.org/debian unstable/contrib Sources.diff/Index [63.3 kB]
            131s Get:4 http://deb.debian.org/debian unstable/main amd64 Packages.diff/Index [63.6 kB]
            131s Get:5 http://deb.debian.org/debian unstable/main Sources T-2025-09-28-2013.43-F-2025-09-28-0810.40.pdiff [69.0 kB]
            131s Get:6 http://deb.debian.org/debian unstable/contrib Sources T-2025-09-28-1412.22-F-2025-09-28-1412.22.pdiff [31 B]
            132s autopkgtest [22:29:56]: test test-suite: [-----------------------
            132s + tmpdir=/tmp/autopkgtest.4eb2Ze/autopkgtest_tmp
            132s + cp -r spec /tmp/autopkgtest.4eb2Ze/autopkgtest_tmp
            132s + cd /tmp/autopkgtest.4eb2Ze/autopkgtest_tmp
            132s + export COVERAGE=no
            140s autopkgtest [22:29:57]: test test-suite:  - - - - - - - - - - results - - - - - - - - - -
            140s test-suite       FAIL
            200s autopkgtest [22:29:58]: test test-suite-skip: preparing testbed
            201s autopkgtest [22:30:56]: test test-suite-skip: [-----------------------
            some lines
            201s test-suite-skip       SKIP
            1079s autopkgtest [22:45:13]: @@@@@@@@@@@@@@@@@@@@ summary
            1079s pybuild-autopkgtest  PASS
            1079s test-suite           FAIL
            1080s test-suite-skip      SKIP
            1081s qemu-system-x86_64: terminating on signal 15 from pid 2264680 (/usr/bin/python3)
        """  # noqa: E501
        ).encode("utf-8")

        self.autopkgtest_log_artifact, _ = playground.create_artifact(
            category=ArtifactCategory.AUTOPKGTEST,
            data=DebianAutopkgtest(
                architecture="amd64",
                cmdline=textwrap.dedent(
                    """\
                    autopkgtest --apt-upgrade --output-dir=artifact-dir
                     --summary=artifact-d/summary --no-built-binaries
                     --needs-internet=run /tmp/debusine-d/phing_3.1.0-1_all.deb
                    /tmp/dir/phing_3.1.0-1.dsc -- incus artifact/2482231 --
                    --profile debusine
                    """
                ),
                distribution="debian:sid",
                results={
                    "help": DebianAutopkgtestResult(
                        details="(superficial)",
                        status=DebianAutopkgtestResultStatus.PASS,
                    ),
                    "php-lint": DebianAutopkgtestResult(
                        details="(superficial)",
                        status=DebianAutopkgtestResultStatus.PASS,
                    ),
                    "phpunit": DebianAutopkgtestResult(
                        details="(superficial)",
                        status=DebianAutopkgtestResultStatus.PASS,
                    ),
                    "version": DebianAutopkgtestResult(
                        details="(superficial)",
                        status=DebianAutopkgtestResultStatus.PASS,
                    ),
                },
                source_package=DebianAutopkgtestSource(
                    name="phing",
                    url=(
                        "https://debusine.debian.net/debian/developers/"
                        "artifact/2489330/download/phing_3.1.0-1.dsc"
                    ),
                    version="3.1.0-1",
                ),
            ),
            paths={"log": self.autopkgtest_log},
            create_files=True,
        )
        autopkgtest_log_file = (
            self.autopkgtest_log_artifact.fileinartifact_set.get(path="log")
        )
        autopkgtest_log_file.content_type = "text/plain; charset=utf-8"
        autopkgtest_log_file.save()

        # Add artifact to the collection
        environment = playground.create_debian_environment().artifact
        assert environment is not None
        work_request = playground.create_sbuild_work_request(
            environment=environment,
            source=playground.create_source_artifact(),
            build_architecture="amd64",
        )
        playground.advance_work_request(work_request, mark_pending=True)

        auotpkgtest_logs_collection.manager.add_artifact(
            self.autopkgtest_log_artifact,
            user=self.user,
            variables={
                "package": "hello",
                "version": "1.0.0-1",
                "architecture": "amd64",
                "timestamp": "10000000",
                "work_request_id": work_request.id,
            },
        )


class DebianWorkspace(DefaultContext):
    """Set up a Debian-like workspace."""

    env_sid_autopkgtest_amd64: CollectionItem
    env_sid_lintian_amd64: CollectionItem

    collection_archive: Collection
    collection_qa_results_sid: Collection
    collection_suite_sid: Collection

    template_upload_to_sid: WorkflowTemplate

    @override
    def build(self, playground: "Playground") -> None:
        super().build(playground)
        self.env_sid_amd64_unshare = playground.create_debian_environment(
            codename="sid",
            architecture="amd64",
            workspace=self.workspace,
        )
        self.env_sid_i386_unshare = playground.create_debian_environment(
            codename="sid",
            architecture="i386",
            workspace=self.workspace,
        )
        self.env_sid_sbuild_amd64 = playground.create_debian_environment(
            codename="sid",
            architecture="amd64",
            variant="sbuild",
            workspace=self.workspace,
        )
        self.env_sid_sbuild_i386 = playground.create_debian_environment(
            codename="sid",
            architecture="i386",
            variant="sbuild",
            workspace=self.workspace,
        )
        self.env_sid_autopkgtest_amd64 = playground.create_debian_environment(
            codename="sid",
            architecture="amd64",
            backend=BackendType.INCUS_LXC,
            variant="autopkgtest",
            workspace=self.workspace,
        )
        self.env_sid_autopkgtest_i386 = playground.create_debian_environment(
            codename="sid",
            architecture="i386",
            backend=BackendType.INCUS_LXC,
            variant="autopkgtest",
            workspace=self.workspace,
        )
        self.env_sid_lintian_amd64 = playground.create_debian_environment(
            codename="sid",
            architecture="amd64",
            variant="lintian",
            workspace=self.workspace,
        )
        self.env_sid_lintian_amd64 = playground.create_debian_environment(
            codename="sid",
            architecture="i386",
            variant="lintian",
            workspace=self.workspace,
        )
        self.collection_archive = self.playground.create_singleton_collection(
            category=CollectionCategory.ARCHIVE,
            workspace=self.workspace,
        )
        self.collection_qa_results_sid = self.playground.create_collection(
            workspace=self.workspace,
            name="sid",
            category=CollectionCategory.QA_RESULTS,
        )
        self.collection_suite_sid = self.playground.create_collection(
            workspace=self.workspace,
            name="sid",
            category=CollectionCategory.SUITE,
        )
        self.template_upload_to_sid = playground.create_workflow_template(
            "upload-to-unstable",
            "debian_pipeline",
            workspace=self.workspace,
            static_parameters={
                "autopkgtest_backend": "incus-lxc",
                "codename": "sid",
                "enable_debdiff": True,
                "enable_reverse_dependencies_autopkgtest": True,
                "enable_upload": True,
                "extra_repositories": [
                    {
                        "components": [
                            "contrib",
                            "non-free",
                            "non-free-firmware",
                        ],
                        "suite": "sid",
                        "url": "http://deb.debian.org/debian",
                    },
                    {
                        "components": [
                            "main",
                            "contrib",
                            "non-free",
                            "non-free-firmware",
                        ],
                        "suite": "buildd-sid",
                        "url": "http://incoming.debian.org/debian-buildd",
                    },
                ],
                "lintian_backend": "unshare",
                "piuparts_backend": "unshare",
                "qa_suite": "sid@debian:suite",
                "regression_tracking_qa_results": "sid@debian:qa-results",
                "sbuild_backend": "unshare",
                "upload_include_binaries": False,
                "upload_merge_uploads": False,
                "vendor": "debian",
            },
            runtime_parameters={
                "arch_all_build_architecture": RuntimeParameter.ANY,
                "architectures": RuntimeParameter.ANY,
                "autopkgtest_backend": ["incus-lxc", "incus-vm"],
                "enable_autopkgtest": RuntimeParameter.ANY,
                "enable_blhc": RuntimeParameter.ANY,
                "enable_confirmation": RuntimeParameter.ANY,
                "enable_debdiff": RuntimeParameter.ANY,
                "enable_lintian": RuntimeParameter.ANY,
                "enable_piuparts": RuntimeParameter.ANY,
                "enable_regression_tracking": RuntimeParameter.ANY,
                "enable_reverse_dependencies_autopkgtest": RuntimeParameter.ANY,
                "enable_upload": RuntimeParameter.ANY,
                "qa_failure_policy": RuntimeParameter.ANY,
                "source_artifact": RuntimeParameter.ANY,
                "upload_delayed_days": RuntimeParameter.ANY,
                "upload_include_binaries": RuntimeParameter.ANY,
                "upload_include_source": RuntimeParameter.ANY,
                "upload_merge_uploads": RuntimeParameter.ANY,
                "upload_require_signature": RuntimeParameter.ANY,
                "upload_since_version": RuntimeParameter.ANY,
            },
        )


class DebianWorkspaceAPI(DefaultContextAPI, DebianWorkspace):
    """DefaultContext, plus a user token."""


class UIRegressionTesting(DebianWorkspace):
    """UI scenario with a regression testing QA workflow."""

    #: Reference version of the sample package
    reference_source: Artifact
    reference_binaries: dict[str, Artifact]

    #: New version of the sample package
    new_source: Artifact

    #: Reverse dependency of the sample package
    dep_source: Artifact
    dep_binaries: dict[str, Artifact]

    #: Root workflow
    workflow_debian_pipeline: WorkRequest

    #: QA subworkflows
    workflow_qa_reference: WorkRequest
    workflow_qa: WorkRequest

    autopkgtest_libfoo_ref_i386: Artifact
    autopkgtest_libfoo_new_amd64: Artifact
    autopkgtest_libfoo_new_i386: Artifact
    autopkgtest_foo_ref_amd64: Artifact
    autopkgtest_foo_ref_i386: Artifact
    autopkgtest_foo_new_amd64: Artifact
    autopkgtest_foo_new_i386: Artifact

    ra_autopkgtest_amd64: RegressionAnalysis
    ra_autopkgtest_i386: RegressionAnalysis
    ra_autopkgtest_foo_amd64: RegressionAnalysis
    ra_autopkgtest_foo_i386: RegressionAnalysis
    ra_lintian_amd64: RegressionAnalysis
    ra_lintian_i386: RegressionAnalysis
    ra_piuparts_amd64: RegressionAnalysis
    ra_piuparts_i386: RegressionAnalysis
    ra_blhc_amd64: RegressionAnalysis
    ra_blhc_i386: RegressionAnalysis

    def __init__(
        self, *, set_current: bool = False, simulate: bool = True
    ) -> None:
        """Store scenario arguments."""
        super().__init__(set_current=set_current)
        self.simulate = simulate

    def _create_sample_package(
        self,
        source_name: str,
        version: str,
        *,
        binary_name: str | None = None,
        dsc_fields: dict[str, Any] | None = None,
        deb_fields: dict[str, Any] | None = None,
    ) -> tuple[Artifact, dict[str, Artifact]]:
        """Create a source package and its binaries, and add them to sid."""
        binary_name = binary_name or source_name
        source = self.playground.create_minimal_source_package_artifact(
            name=source_name,
            version=version,
            dsc_fields={"Binary": binary_name, **(dsc_fields or {})},
        )
        self.collection_suite_sid.manager.add_artifact(
            source,
            user=self.user,
            variables={"component": "main", "section": "devel"},
        )
        binaries = {
            arch: self.playground.create_minimal_binary_package_artifact(
                srcpkg_name=source_name,
                srcpkg_version=version,
                package=binary_name,
                version=version,
                architecture=arch,
                deb_fields=deb_fields,
            )
            for arch in ("amd64", "i386")
        }
        for binary in binaries.values():
            self.collection_suite_sid.manager.add_artifact(
                binary,
                user=self.user,
                variables={
                    "component": "main",
                    "section": "misc",
                    "priority": "optional",
                },
            )
        for arch in binaries:
            build_log = self.playground.create_build_log_artifact(
                source=source_name, version=version, build_arch=arch
            )
            self.playground.create_artifact_relation(
                artifact=build_log, target=binaries[arch]
            )
        return source, binaries

    def _add_regression_analysis(self, **analysis: RegressionAnalysis) -> None:
        od = self.workflow_qa.output_data
        assert od is not None
        assert od.regression_analysis is not None
        for name, a in analysis.items():
            od.regression_analysis[name] = a
        self.workflow_qa.output_data = od
        self.workflow_qa.save()

    def set_work_request_result(
        self, work_request: WorkRequest, result: WorkRequest.Results
    ) -> None:
        """Simulate that the work request completed with the given result."""
        work_request.configured_task_data = work_request.task_data
        work_request.save()
        self.playground.advance_work_request(work_request, result=result)

    def simulate_work_request_result(
        self,
        artifact: Artifact,
        work_request: WorkRequest,
        result: WorkRequest.Results = WorkRequest.Results.SUCCESS,
    ) -> None:
        """
        Generate an artifact as a result for a work request.

        The artifact is then added to collections according to the work request
        UPDATE_COLLECTION_WITH_ARTIFACTS success hooks.
        """
        artifact.created_by_work_request = work_request
        artifact.save()
        self.set_work_request_result(work_request, result)

    def add_result_autopkgtest(
        self,
        package: Artifact,
        architecture: str,
        *,
        parent: WorkRequest,
        result: WorkRequest.Results = WorkRequest.Results.SUCCESS,
        status: DebianAutopkgtestResultStatus = (
            DebianAutopkgtestResultStatus.PASS
        ),
    ) -> Artifact:
        """Add an Autopkgtest QA artifact to collection_qa_results_sid."""
        work_request = parent.children.get(
            task_type=TaskTypes.WORKER,
            task_name="autopkgtest",
            task_data__build_architecture=architecture,
        )
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.AUTOPKGTEST,
            data=DebianAutopkgtest(
                results={"test": DebianAutopkgtestResult(status=status)},
                cmdline="autopkgtest --this-is-a-test",
                source_package=DebianAutopkgtestSource(
                    name=package.data["name"],
                    version=package.data["version"],
                    url="https://example.org/test",
                ),
                architecture=architecture,
                distribution="debian",
            ),
        )
        self.simulate_work_request_result(artifact, work_request, result)
        return artifact

    def add_result_lintian(
        self,
        package: Artifact,  # noqa: ARG002, U100
        architecture: str,
        *,
        parent: WorkRequest,
        result: WorkRequest.Results = WorkRequest.Results.SUCCESS,
    ) -> Artifact:
        """Add a Lintian QA result artifact to collection_qa_results_sid."""
        work_request = parent.children.get(
            task_type=TaskTypes.WORKER,
            task_name="lintian",
            task_data__build_architecture=architecture,
        )
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.LINTIAN,
            data=DebianLintian(
                architecture=architecture,
                summary=DebianLintianSummary(
                    tags_count_by_severity={},
                    package_filename={},
                    tags_found=[],
                    overridden_tags_found=[],
                    lintian_version="3.14",
                    distribution="sid",
                ),
            ),
        )
        self.simulate_work_request_result(artifact, work_request, result)
        return artifact

    def add_result_piuparts(
        self,
        package: Artifact,  # noqa: ARG002, U100
        architecture: str,
        *,
        parent: WorkRequest,
        result: WorkRequest.Results = WorkRequest.Results.SUCCESS,
    ) -> Artifact:
        """Add a Piuparts QA result artifact to collection_qa_results_sid."""
        work_request = parent.children.get(
            task_type=TaskTypes.WORKER,
            task_name="piuparts",
            task_data__build_architecture=architecture,
        )
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.PIUPARTS,
            data=DebianPiuparts(
                piuparts_version="3.14",
                binary_packages={},
                architecture=architecture,
                distribution="debian",
            ),
        )
        self.simulate_work_request_result(artifact, work_request, result)
        return artifact

    def add_result_blhc(
        self,
        package: Artifact,
        architecture: str,
        parent: WorkRequest,
        result: WorkRequest.Results = WorkRequest.Results.SUCCESS,
    ) -> Artifact:
        """Add a blhc QA result artifact to collection_qa_results_sid."""
        work_request = parent.children.get(
            task_type=TaskTypes.WORKER,
            task_name="blhc",
            task_data__build_architecture=architecture,
        )
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.BLHC,
            data=DebianBlhc(
                source=package.data["name"],
                version=package.data["version"],
                architecture=architecture,
            ),
        )
        self.simulate_work_request_result(artifact, work_request, result)
        return artifact

    def simulate_sbuild(self) -> None:
        """Simulate results of sbuild workflow."""
        workflow_sbuild = self.workflow_debian_pipeline.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        for sbuild in workflow_sbuild.children.filter(
            task_type=TaskTypes.WORKER, task_name="sbuild"
        ):
            assert sbuild.workflow_data.step is not None
            arch = sbuild.workflow_data.step.removeprefix("build-")
            upload = self.playground.create_upload_artifacts(
                src_name=self.new_source.data["name"],
                version=self.new_source.data["version"],
                binary=True,
                source=False,
                binaries=[(self.new_source.data["name"], arch)],
                workspace=sbuild.workspace,
                work_request=sbuild,
            )
            build_log = self.playground.create_build_log_artifact(
                source=self.new_source.data["name"],
                version=self.new_source.data["version"],
                build_arch=arch,
                workspace=sbuild.workspace,
                work_request=sbuild,
            )
            self.playground.create_artifact_relation(
                artifact=build_log, target=upload.upload
            )
            self.set_work_request_result(sbuild, WorkRequest.Results.SUCCESS)

    def simulate_autopkgtest(self) -> None:
        """Simulate results of autopkgtest workflow."""
        # Results from a previous run
        workflow_autopkgtest_reference = (
            self.workflow_qa_reference.children.get(
                task_type=TaskTypes.WORKFLOW, task_name="autopkgtest"
            )
        )
        # No amd64 result
        self.autopkgtest_libfoo_ref_i386 = self.add_result_autopkgtest(
            self.reference_source, "i386", parent=workflow_autopkgtest_reference
        )

        # New run
        workflow_autopkgtest = self.workflow_qa.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="autopkgtest"
        )
        self.playground.advance_work_request(
            workflow_autopkgtest,
            mark_running=True,
            result=WorkRequest.Results.FAILURE,
        )
        self.autopkgtest_libfoo_new_amd64 = self.add_result_autopkgtest(
            self.new_source, "amd64", parent=workflow_autopkgtest
        )
        self.autopkgtest_libfoo_new_i386 = self.add_result_autopkgtest(
            self.new_source, "i386", parent=workflow_autopkgtest
        )
        self.ra_autopkgtest_amd64 = RegressionAnalysis(
            original_source_version=None,
            original_url=None,
            new_source_version=self.new_source.data["version"],
            new_url=f"/{self.workspace}/artifact/10000/",
            status=RegressionAnalysisStatus.NO_RESULT,
            details={"build": "regression"},
            original_artifact_id=None,
            new_artifact_id=self.autopkgtest_libfoo_new_amd64.id,
        )
        self.ra_autopkgtest_i386 = RegressionAnalysis(
            original_source_version=self.reference_source.data["version"],
            original_url=f"/{self.workspace}/artifact/10001/",
            new_source_version=self.new_source.data["version"],
            new_url="f/{self.workspace}/artifact/10002/",
            status=RegressionAnalysisStatus.STABLE,
            details={"build": "stable"},
            original_artifact_id=self.autopkgtest_libfoo_ref_i386.id,
            new_artifact_id=self.autopkgtest_libfoo_new_i386.id,
        )
        workflow_autopkgtest.output_data = OutputData(
            regression_analysis={
                "amd64": self.ra_autopkgtest_amd64,
                "i386": self.ra_autopkgtest_i386,
            }
        )
        workflow_autopkgtest.save()

        self._add_regression_analysis(
            **{
                f"autopkgtest:{self.new_source.data['name']}:amd64": (
                    self.ra_autopkgtest_amd64
                ),
                f"autopkgtest:{self.new_source.data['name']}:i386": (
                    self.ra_autopkgtest_i386
                ),
            }
        )

    def simulate_rev_dep_autopkgtest(self) -> None:
        """Simulate autopkgtest of reverse dependency."""
        # Results from a previous run
        workflow_autopkgtest_foo_reference = (
            self.workflow_qa_reference.children.get(
                task_type=TaskTypes.WORKFLOW,
                task_name="reverse_dependencies_autopkgtest",
            ).children.get()
        )
        self.autopkgtest_foo_ref_amd64 = self.add_result_autopkgtest(
            self.dep_source, "amd64", parent=workflow_autopkgtest_foo_reference
        )
        self.autopkgtest_foo_ref_i386 = self.add_result_autopkgtest(
            self.dep_source,
            "i386",
            result=WorkRequest.Results.FAILURE,
            parent=workflow_autopkgtest_foo_reference,
            status=DebianAutopkgtestResultStatus.FAIL,
        )

        # New run
        workflow_autopkgtest_foo = self.workflow_qa.children.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="reverse_dependencies_autopkgtest",
        ).children.get()
        self.playground.advance_work_request(
            workflow_autopkgtest_foo,
            mark_running=True,
            result=WorkRequest.Results.SUCCESS,
        )
        self.autopkgtest_foo_new_amd64 = self.add_result_autopkgtest(
            self.dep_source, "amd64", parent=workflow_autopkgtest_foo
        )
        self.autopkgtest_foo_new_i386 = self.add_result_autopkgtest(
            self.dep_source, "i386", parent=workflow_autopkgtest_foo
        )
        self.ra_autopkgtest_foo_amd64 = RegressionAnalysis(
            original_source_version=self.dep_source.data["version"],
            original_url=f"/{self.workspace}/artifact/10010/",
            new_source_version=self.dep_source.data["version"],
            new_url=f"/{self.workspace}/artifact/10011/",
            status=RegressionAnalysisStatus.STABLE,
            details={"build": "stable"},
            original_artifact_id=self.autopkgtest_foo_ref_amd64.id,
            new_artifact_id=self.autopkgtest_foo_new_amd64.id,
        )
        self.ra_autopkgtest_foo_i386 = RegressionAnalysis(
            original_source_version=self.dep_source.data["version"],
            original_url=f"/{self.workspace}/artifact/10012/",
            new_source_version=self.dep_source.data["version"],
            new_url="f/{self.workspace}/artifact/10013/",
            status=RegressionAnalysisStatus.IMPROVEMENT,
            details={"build": "improvement"},
            original_artifact_id=self.autopkgtest_foo_ref_i386.id,
            new_artifact_id=self.autopkgtest_foo_new_i386.id,
        )
        workflow_autopkgtest_foo.output_data = OutputData(
            regression_analysis={
                "amd64": self.ra_autopkgtest_foo_amd64,
                "i386": self.ra_autopkgtest_foo_i386,
            }
        )
        workflow_autopkgtest_foo.save()

        self._add_regression_analysis(
            **{
                f"autopkgtest:{self.dep_source.data['name']}:amd64": (
                    self.ra_autopkgtest_foo_amd64
                ),
                f"autopkgtest:{self.dep_source.data['name']}:i386": (
                    self.ra_autopkgtest_foo_i386
                ),
            }
        )

    def simulate_lintian(self) -> None:
        """Simulate lintian."""
        # Results from a previous run
        workflow_lintian_reference = self.workflow_qa_reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="lintian"
        )
        ref_amd64 = self.add_result_lintian(
            self.reference_source, "amd64", parent=workflow_lintian_reference
        )
        ref_i386 = self.add_result_lintian(
            self.reference_source,
            "i386",
            result=WorkRequest.Results.FAILURE,
            parent=workflow_lintian_reference,
        )

        # New run
        workflow_lintian = self.workflow_qa.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="lintian"
        )
        self.playground.advance_work_request(
            workflow_lintian,
            mark_running=True,
            result=WorkRequest.Results.FAILURE,
        )
        new_amd64 = self.add_result_lintian(
            self.new_source, "amd64", parent=workflow_lintian
        )
        new_i386 = self.add_result_lintian(
            self.new_source, "i386", parent=workflow_lintian
        )
        self.ra_lintian_amd64 = RegressionAnalysis(
            original_source_version=self.new_source.data["version"],
            original_url=f"/{self.workspace}/artifact/10022/",
            new_source_version=self.new_source.data["version"],
            new_url=f"/{self.workspace}/artifact/10023/",
            status=RegressionAnalysisStatus.STABLE,
            details={"added_tags": [], "removed_tags": []},
            original_artifact_id=ref_amd64.id,
            new_artifact_id=new_amd64.id,
        )
        self.ra_lintian_i386 = RegressionAnalysis(
            original_source_version=self.new_source.data["version"],
            original_url=f"/{self.workspace}/artifact/10024/",
            new_source_version=self.new_source.data["version"],
            new_url=f"/{self.workspace}/artifact/10025/",
            status=RegressionAnalysisStatus.IMPROVEMENT,
            details={"added_tags": [], "removed_tags": ["no-nmu-in-changelog"]},
            original_artifact_id=ref_i386.id,
            new_artifact_id=new_i386.id,
        )
        workflow_lintian.output_data = OutputData(
            regression_analysis={
                "amd64": self.ra_lintian_amd64,
                "i386": self.ra_lintian_i386,
            }
        )
        workflow_lintian.save()

        self._add_regression_analysis(
            **{
                f"lintian:{self.new_source.data['name']}:amd64": (
                    self.ra_lintian_amd64
                ),
                f"lintian:{self.new_source.data['name']}:i386": (
                    self.ra_lintian_i386
                ),
            }
        )

    def simulate_piuparts(self) -> None:
        """Simulate piuparts."""
        # Results from a previous run
        workflow_piuparts_reference = self.workflow_qa_reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="piuparts"
        )
        ref_amd64 = self.add_result_piuparts(
            self.reference_source, "amd64", parent=workflow_piuparts_reference
        )
        ref_i386 = self.add_result_piuparts(
            self.reference_source,
            "i386",
            result=WorkRequest.Results.FAILURE,
            parent=workflow_piuparts_reference,
        )

        # New run
        workflow_piuparts = self.workflow_qa.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="piuparts"
        )
        self.playground.advance_work_request(
            workflow_piuparts,
            mark_running=True,
            result=WorkRequest.Results.SUCCESS,
        )
        new_amd64 = self.add_result_piuparts(
            self.new_source, "amd64", parent=workflow_piuparts
        )
        new_i386 = self.add_result_piuparts(
            self.new_source, "i386", parent=workflow_piuparts
        )
        self.ra_piuparts_amd64 = RegressionAnalysis(
            original_source_version=self.new_source.data["version"],
            original_url=f"/{self.workspace}/artifact/10030/",
            new_source_version=self.new_source.data["version"],
            new_url=f"/{self.workspace}/artifact/10031/",
            status=RegressionAnalysisStatus.STABLE,
            original_artifact_id=ref_amd64.id,
            new_artifact_id=new_amd64.id,
        )
        self.ra_piuparts_i386 = RegressionAnalysis(
            original_source_version=self.new_source.data["version"],
            original_url=f"/{self.workspace}/artifact/10032/",
            new_source_version=self.new_source.data["version"],
            new_url=f"/{self.workspace}/artifact/10033/",
            status=RegressionAnalysisStatus.IMPROVEMENT,
            original_artifact_id=ref_i386.id,
            new_artifact_id=new_i386.id,
        )
        workflow_piuparts.output_data = OutputData(
            regression_analysis={
                "amd64": self.ra_piuparts_amd64,
                "i386": self.ra_piuparts_i386,
            }
        )
        workflow_piuparts.save()

        self._add_regression_analysis(
            **{
                f"piuparts:{self.new_source.data['name']}:amd64": (
                    self.ra_piuparts_amd64
                ),
                f"piuparts:{self.new_source.data['name']}:i386": (
                    self.ra_piuparts_i386
                ),
            }
        )

    def simulate_blhc(self) -> None:
        """Simulate blhc."""
        # Results from a previous run
        workflow_blhc_reference = self.workflow_qa_reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="blhc"
        )
        ref_amd64 = self.add_result_blhc(
            self.reference_source, "amd64", parent=workflow_blhc_reference
        )
        ref_i386 = self.add_result_blhc(
            self.reference_source, "i386", parent=workflow_blhc_reference
        )

        # New run
        workflow_blhc = self.workflow_qa.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="blhc"
        )
        self.playground.advance_work_request(
            workflow_blhc,
            mark_running=True,
            result=WorkRequest.Results.FAILURE,
        )
        new_amd64 = self.add_result_blhc(
            self.reference_source, "amd64", parent=workflow_blhc
        )
        new_i386 = self.add_result_blhc(
            self.reference_source,
            "i386",
            parent=workflow_blhc,
            result=WorkRequest.Results.FAILURE,
        )
        self.ra_blhc_amd64 = RegressionAnalysis(
            original_source_version=self.new_source.data["version"],
            original_url=f"/{self.workspace}/artifact/10040/",
            new_source_version=self.new_source.data["version"],
            new_url=f"/{self.workspace}/artifact/10041/",
            status=RegressionAnalysisStatus.STABLE,
            original_artifact_id=ref_amd64.id,
            new_artifact_id=new_amd64.id,
        )
        self.ra_blhc_i386 = RegressionAnalysis(
            original_source_version=self.new_source.data["version"],
            original_url=f"/{self.workspace}/artifact/10042/",
            new_source_version=self.new_source.data["version"],
            new_url=f"/{self.workspace}/artifact/10043/",
            status=RegressionAnalysisStatus.REGRESSION,
            original_artifact_id=ref_i386.id,
            new_artifact_id=new_i386.id,
        )
        workflow_blhc.output_data = OutputData(
            regression_analysis={
                "amd64": self.ra_blhc_amd64,
                "i386": self.ra_blhc_i386,
            }
        )
        workflow_blhc.save()

        self._add_regression_analysis(
            **{
                f"blhc:{self.new_source.data['name']}:amd64": (
                    self.ra_blhc_amd64
                ),
                f"blhc:{self.new_source.data['name']}:i386": (
                    self.ra_blhc_i386
                ),
            }
        )

    @override
    def build(self, playground: "Playground") -> None:
        super().build(playground)

        # Create reference version of package under test.
        self.reference_source, self.reference_binaries = (
            self._create_sample_package(
                "libfoo", "1.0-1", binary_name="libfoo1"
            )
        )

        # Create a package that depends on the package under test.
        self.dep_source, self.dep_binaries = self._create_sample_package(
            "foo",
            "2.0-1",
            dsc_fields={"Testsuite": "autopkgtest"},
            deb_fields={"Depends": "libfoo1"},
        )

        # Create new version of package under test.
        self.new_source = playground.create_source_artifact(
            name="libfoo", binaries=["libfoo1"], version="1.0-2"
        )

        # Root workflow.
        self.workflow_debian_pipeline = playground.create_workflow(
            self.template_upload_to_sid,
            task_data={
                "source_artifact": self.new_source.pk,
                "architectures": ["all", "amd64", "i386"],
                "enable_regression_tracking": True,
            },
            workspace=self.workspace,
        )
        assert (
            self.workflow_debian_pipeline.status == WorkRequest.Statuses.PENDING
        )
        assert orchestrate_workflow(self.workflow_debian_pipeline)

        # QA workflows.
        self.workflow_qa_reference = self.workflow_debian_pipeline.children.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="qa",
            task_data__update_qa_results=RegressionTrackingUpdateQAResults.YES,
        )
        self.workflow_qa = self.workflow_debian_pipeline.children.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="qa",
            task_data__update_qa_results=RegressionTrackingUpdateQAResults.NO,
        )
        self.workflow_qa.output_data = OutputData(
            regression_analysis={
                "": RegressionAnalysis(
                    original_source_version=self.reference_source.data[
                        "version"
                    ],
                    new_source_version=self.new_source.data["version"],
                    status=RegressionAnalysisStatus.REGRESSION,
                    original_artifact_id=self.reference_source.id,
                    new_artifact_id=self.new_source.id,
                ),
            }
        )
        self.workflow_qa.save()

        # Simulate workflow results
        if self.simulate:
            self.simulate_sbuild()
            self.simulate_autopkgtest()
            self.simulate_rev_dep_autopkgtest()
            self.simulate_lintian()
            self.simulate_piuparts()

            # Sbuild has completed so blhc can be populated
            context.clear_lookup_cache()
            orchestrate_workflow(self.workflow_debian_pipeline)
            self.simulate_blhc()


class UIPlayground(PackageBuildLog, UIRegressionTesting):
    """
    Base scenario for UI tests.

    This gives a default password to the test user.
    """

    suite: Collection
    env_amd64: Artifact
    env_s390x: Artifact
    source_hello: Artifact
    source_dpkg: Artifact
    source_udev: Artifact
    template_sbuild: WorkflowTemplate
    template_noop: WorkflowTemplate
    worker_pool: WorkerPool
    worker_static: Worker
    worker_celery: Worker
    worker_in_pool: Worker
    worker_in_pool_busy: Worker

    @override
    def build(self, playground: "Playground") -> None:
        super().build(playground)
        self.set_current()

        # Set a password for the test user
        self.user.set_password("playground")
        self.user.save()

        asset = self.playground.create_signing_key_asset()
        self.playground.create_asset_usage(asset)

        # Pretend the user comes from Salsa
        Identity.objects.create(
            user=self.user,
            issuer="salsa",
            subject="playground@debian.example.org",
            claims={"test": True},
        )

        other_user = User.objects.create_user(
            username="playground-other", first_name="Other", last_name="User"
        )

        # Make them owners of the playground workspace...
        role_group = self.playground.create_group_role(
            self.workspace,
            Workspace.Roles.OWNER,
            users=[self.user],
            name="Admins",
        )
        # ...and admin of the owners group
        role_group.set_user_role(self.user, Group.Roles.ADMIN)

        # Create a second group with both users in it
        other_group = Group.objects.create(name="Playground", scope=self.scope)
        other_group.add_user(self.user)
        other_group.add_user(other_user)

        # Create an unscoped Admins group
        ws_group = Group.objects.create(name="Admins", scope=self.scope)
        ws_group.add_user(self.user)
        ws_group.set_user_role(self.user, Group.Roles.ADMIN)

        # Create a worker pool and some workers
        self.worker_pool = self.playground.create_worker_pool("playground")

        self.worker_static = self.playground.create_worker(
            fqdn="playground.lan"
        )
        self.worker_celery = self.playground.create_worker(
            worker_type=WorkerType.CELERY, fqdn="playground.lan"
        )
        self.worker_in_pool = self.playground.create_worker(
            fqdn="playground.lan", worker_pool=self.worker_pool
        )
        self.worker_in_pool_busy = self.playground.create_worker(
            fqdn="playground.lan", worker_pool=self.worker_pool
        )

        # Create a Debian scope, to test multi-scope UI elements
        self.playground.get_or_create_scope(
            "debian", label="Debian", icon="web/icons/debian-openlogo-nd.svg"
        )

        # Create a sbuild workflow template
        self.template_sbuild = self.playground.create_workflow_template(
            name="Build package", task_name="sbuild", static_parameters={}
        )

        # Create a debian:suite collection
        self.suite = self.playground.create_collection(
            workspace=self.workspace,
            name="play_bookworm",
            category=CollectionCategory.SUITE,
            data={
                "may_reuse_versions": False,
                "release_fields": {
                    "Suite": "stable",
                    "Codename": "bookworm",
                },
                "components": [
                    "main",
                    "contrib",
                    "non-free-firmware",
                    "non-free",
                ],
                "architectures": [
                    "all",
                    "amd64",
                    "arm64",
                    "armel",
                    "armhf",
                    "i386",
                    "mips64el",
                    "mipsel",
                    "ppc64el",
                    "s390x",
                ],
            },
        )

        # create another debian:suite collection
        trixie = self.playground.create_collection(
            workspace=self.workspace,
            name="play_trixie",
            category=CollectionCategory.SUITE,
            data={
                "may_reuse_versions": False,
                "release_fields": {
                    "Suite": "stable",
                    "Codename": "trixie",
                },
                "components": [
                    "main",
                    "contrib",
                    "non-free-firmware",
                    "non-free",
                ],
                "architectures": [
                    "all",
                    "amd64",
                    "arm64",
                    "armel",
                    "armhf",
                    "i386",
                    "mips64el",
                    "mipsel",
                    "ppc64el",
                    "s390x",
                ],
            },
        )

        # Add collection relations
        CollectionRelation.objects.create(
            source=trixie,
            target=self.suite,
            created_by_user=self.user,
            type=CollectionRelation.Types.SUITE_BASED_ON,
        )

        # Add collection relation requires to see the re-ordering
        # create another debian:suite collection
        personal_suite = self.playground.create_collection(
            workspace=self.workspace,
            name="r-personal",
            category=CollectionCategory.SUITE,
        )

        trixie_backports_suite = self.playground.create_collection(
            workspace=self.workspace,
            name="trixie-backports",
            category=CollectionCategory.SUITE,
        )
        trixie_security_suite = self.playground.create_collection(
            workspace=self.workspace,
            name="trixie-security",
            category=CollectionCategory.SUITE,
        )

        CollectionRelation.objects.create(
            source=personal_suite,
            target=trixie,
            created_by_user=self.user,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=1,
        )
        CollectionRelation.objects.create(
            source=personal_suite,
            target=trixie_backports_suite,
            created_by_user=self.user,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=2,
        )
        CollectionRelation.objects.create(
            source=personal_suite,
            target=trixie_security_suite,
            created_by_user=self.user,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=3,
        )

        # Create debian environments to simulate builds
        item = self.playground.create_debian_environment(architecture="amd64")
        assert item.artifact is not None
        self.env_amd64 = item.artifact
        item = self.playground.create_debian_environment(architecture="s390x")
        assert item.artifact is not None
        self.env_s390x = item.artifact

        # Create the source packages
        self.source_hello = self.playground.create_source_artifact(
            name="hello", version="1.0-1", create_files=True
        )
        self.source_dpkg = self.playground.create_source_artifact(
            name="dpkg",
            binaries=["dpkg", "dpkg-doc", "libdpkg-perl"],
            version="1.21.22",
            create_files=True,
        )
        self.source_udev = self.playground.create_source_artifact(
            name="udev", version="252.26-1~deb12u2", create_files=True
        )

        # Populate the debian:suite collection with artifacts
        wr = self.playground.simulate_package_build(
            self.source_hello, architecture="amd64", worker=self.worker_static
        )
        self.add_results_to_suite(wr)
        wr = self.playground.simulate_package_build(
            self.source_dpkg,
            binaries=["dpkg"],
            architecture="amd64",
            worker=self.worker_static,
        )
        self.add_results_to_suite(wr)
        wr = self.playground.simulate_package_build(
            self.source_dpkg,
            binaries=["dpkg"],
            architecture="armhf",
            worker=self.worker_static,
        )
        self.add_results_to_suite(wr)
        wr = self.playground.simulate_package_build(
            self.source_dpkg,
            binaries=["dpkg-doc", "libdpkg-perl"],
            architecture="all",
            build_architecture="amd64",
            worker=self.worker_static,
        )
        self.add_results_to_suite(wr)

        # Create a variety of work requests
        wr = self.playground.create_worker_task(
            task_name="noop", mark_pending=True
        )
        wr.task_data = {"result": WorkRequest.Results.FAILURE}
        wr.configured_task_data = {"result": WorkRequest.Results.SUCCESS}
        self.playground.advance_work_request(
            wr, result=WorkRequest.Results.SUCCESS
        )
        self.playground.create_worker_task(
            task_name="noop", result=WorkRequest.Results.FAILURE
        )
        self.playground.create_worker_task(
            task_name="noop", result=WorkRequest.Results.ERROR
        )
        self.playground.create_worker_task(task_name="noop", mark_aborted=True)

        wr = self.playground.create_worker_task(
            task_name="noop", worker=self.worker_in_pool_busy, mark_running=True
        )

        self.simulate_sbuild_workflow(self.template_sbuild, self.source_udev)

        # Create a binary DebDiff artifact and its required relation
        debdiff_report = textwrap.dedent(
            """
            some debdiff text header"

            Files in second .deb but not in first
            -------------------------------------
            -rw-r--r--  root/root   /etc/simplemonitor/monitor.ini
            -rw-r--r--  root/root   DEBIAN/conffiles

            Files in first .deb but not in second
            ------------------------------------
            -rw-r--r--  root/root   /usr/lib/python3/dist-packages/pyaarlo
            -rw-r--r--  root/root   /usr/share/doc/python3-pyaarlo/READ.gz

            Control files: lines which differ (wdiff format)
            ------------------------------------------------
            Description: [-one description-] {+another description+}
            Homepage: [-https://github.com/twrecked/pyaarlo-]
            {+https://simplemonitor.readthedocs.io+}"""
        ).encode("utf-8")

        artifact, _ = playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: debdiff_report},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={"original": "hello-1.deb", "new": "hello-2.deb"},
        )

        target, _ = playground.create_artifact(
            category=ArtifactCategory.UPLOAD,
            data=DebianUpload(
                type="dpkg",
                changes_fields={
                    "Architecture": "amd64",
                    "Files": [{"name": "hello_1.0_amd64.deb"}],
                },
            ),
        )

        ArtifactRelation.objects.create(
            artifact=artifact,
            target=target,
            type=ArtifactRelation.Relations.RELATES_TO,
        )

        # Create a source DebDiff artifact and its required relation
        debdiff_report = textwrap.dedent(
            """\
            diff -u -N original/added.txt new/added.txt
            --- original/added.txt	1970-01-01 01:00:00.000000000 +0100
            +++ new/added.txt	2025-03-26 12:42:27.672906377 +0000
            @@ -0,0 +1 @@
            +new file
            diff -u -N original/changed.txt new/changed.txt
            --- original/changed.txt	2025-03-26 12:41:44.191924503 +0000
            +++ new/changed.txt	2025-03-26 12:42:21.945798421 +0000
            @@ -1,3 +1,2 @@
             1
             2
            -3
            diff -u -N original/removed.txt new/removed.txt
            --- original/removed.txt	2025-03-26 12:42:37.607381416 +0000
            +++ new/removed.txt	1970-01-01 01:00:00.000000000 +0100
            @@ -1 +0,0 @@
            -removed file
        """
        ).encode("utf-8")

        artifact, _ = self.playground.create_artifact(
            paths={DebDiff.CAPTURE_OUTPUT_FILENAME: debdiff_report},
            category=ArtifactCategory.DEBDIFF,
            create_files=True,
            data={"original": "hello-1.dsc", "new": "hello-2.dsc"},
        )

        target, _ = playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data=DebianSourcePackage(
                name="hello",
                version="1.0",
                type="dpkg",
                dsc_fields={},
            ),
        )
        ArtifactRelation.objects.create(
            artifact=artifact,
            target=target,
            type=ArtifactRelation.Relations.RELATES_TO,
        )

        # Create a variety of workflows

        self.template_noop = playground.create_workflow_template(
            name="Dummy",
            task_name="noop",
        )

        # 3 completed workflow
        for _ in range(3):
            self.playground.create_workflow(
                task_name=self.template_noop, result=WorkRequest.Results.SUCCESS
            )

        # 2 running workflows
        for _ in range(2):
            self.playground.create_workflow(
                task_name=self.template_noop, mark_running=True
            )

        # 1 needs input workflow
        workflow_needs_input = self.playground.create_workflow(
            task_name=self.template_noop, mark_running=True
        )
        workflow_needs_input.workflow_runtime_status = (
            WorkRequest.RuntimeStatuses.NEEDS_INPUT
        )
        workflow_needs_input.save()

        # Create and populate the default task-configuration collection
        task_config_collection = self.default_task_configuration_collection
        task_config_collection.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.user,
            data={
                "template": "reduce-parallelism",
                "override_values": {"build_options": "parallel=2"},
                "comment": "reduce parallelism to avoid crashing workers",
            },
        )
        task_config_collection.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.user,
            data={
                "task_type": "Worker",
                "task_name": "sbuild",
                "subject": "openjdk-17",
                "use_templates": ["reduce-parallelism"],
                "comment": "reduce parallelism for openjdk-17",
            },
        )
        task_config_collection.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.user,
            data={
                "task_type": "Workflow",
                "task_name": "debian_pipeline",
                "subject": "u-boot",
                "default_values": {
                    "architectures_allowlist": ["arm64", "armhf"],
                },
                "comment": "only build on ARM by default",
            },
        )

    def add_results_to_suite(
        self,
        work_request: WorkRequest,
    ) -> None:
        """
        Add the work request artifacts to the debian:suite collection.

        :param suite: the target suite, or self.suite by default
        """
        suite = self.suite

        source = Artifact.objects.get(
            pk=work_request.task_data["input"]["source_artifact"]
        )

        if not suite.child_items.active().filter(artifact=source).exists():
            suite.manager.add_artifact(
                source,
                user=self.user,
                variables={"component": "main", "section": "devel"},
            )
        for binary in Artifact.objects.filter(
            created_by_work_request=work_request,
            category=ArtifactCategory.BINARY_PACKAGE,
        ):
            suite.manager.add_artifact(
                binary,
                user=self.user,
                variables={
                    "component": "main",
                    "section": "devel",
                    "priority": "optional",
                },
            )

    @context.disable_permission_checks()
    def simulate_sbuild_workflow(
        self,
        template: WorkflowTemplate,
        source: Artifact,
    ) -> WorkRequest:
        """Simulate a sbuild workflow."""
        workflow = self.playground.create_workflow(
            template,
            task_data=SbuildWorkflowData(
                input=SbuildInput(source_artifact=source.pk),
                backend=BackendType.UNSHARE,
                target_distribution="debian:bookworm",
                architectures=["all", "amd64", "s390x"],
            ),
            mark_running=True,
        )

        # A successful build
        self.playground.simulate_package_build(
            source,
            workflow=workflow,
            architecture="amd64",
            worker=self.worker_static,
        )

        # A failed build
        wr_s390x = self.playground.create_sbuild_work_request(
            source=source,
            architecture="s390x",
            environment=self.env_s390x,
            workflow=workflow,
        )
        self.playground.advance_work_request(
            wr_s390x, result=WorkRequest.Results.FAILURE
        )

        # Retrying succeeded
        new_wr_s390x = wr_s390x.retry()
        self.playground.advance_work_request(
            new_wr_s390x, result=WorkRequest.Results.SUCCESS
        )

        return workflow



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the Autopkgtest class."""

import copy
import datetime as dt
import shlex
import textwrap
from pathlib import Path
from typing import cast, override
from unittest.mock import MagicMock, call, create_autospec, patch

import pydantic

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts.local_artifact import AutopkgtestArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianAutopkgtest,
    DebianAutopkgtestResult,
    DebianAutopkgtestSource,
    DebianBinaryPackage,
    DebianSourcePackage,
    DebianUpload,
    WorkRequestResults,
)
from debusine.artifacts.models import DebianAutopkgtestResultStatus as rs
from debusine.client.debusine import Debusine
from debusine.client.models import FileResponse, FilesResponseType, RelationType
from debusine.tasks import Autopkgtest, TaskConfigError
from debusine.tasks.executors import UnshareExecutor
from debusine.tasks.models import (
    AutopkgtestDynamicData,
    AutopkgtestNeedsInternet,
    AutopkgtestTimeout,
    BackendType,
    BaseTaskInputArtifact,
    ExtraExternalRepository,
    LookupMultiple,
    WorkerType,
)
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_artifact_response,
    create_remote_artifact,
    create_system_tarball_data,
)


class AutopkgtestTests(ExternalTaskHelperMixin[Autopkgtest], TestCase):
    """Tests for Autopkgtest."""

    SAMPLE_TASK_DATA = {
        "input": {
            "source_artifact": 5,
            "binary_artifacts": [10, 11],
        },
        "build_architecture": "amd64",
        "environment": "debian/match:codename=bookworm",
    }

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver.set(extra_repositories=None)
        self.resolver.set_field_artifact(
            "environment",
            id=1,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(),
        )
        self.resolver.set_field_artifact(
            "input_source_artifact",
            id=5,
            category=ArtifactCategory.SOURCE_PACKAGE,
            data=DebianSourcePackage(
                name="hello", version="1.0", type="dpkg", dsc_fields={}
            ),
        )
        self.resolver.set_field_artifacts(
            "input_binary_artifacts",
            BaseTaskInputArtifact(
                artifact_id=10,
                category=ArtifactCategory.BINARY_PACKAGE,
                data=self.create_binary_package_data("hello"),
            ),
            BaseTaskInputArtifact(
                artifact_id=11,
                category=ArtifactCategory.UPLOAD,
                data=self.create_upload_data("hello"),
            ),
        )
        self.resolver.set_field_artifacts("input_context_artifacts")

    def mock_executor(self) -> None:
        """Set up a fake unshare executor."""
        mock = create_autospec(UnshareExecutor, instance=True)
        executor = cast(UnshareExecutor, mock)
        self.task.executor = executor
        executor.system_image = self.fake_system_tarball_artifact()
        mock.autopkgtest_virt_server.return_value = "unshare"

    def assert_parse_summary_file_expected(
        self,
        autopkgtest_summary: str,
        expected: dict[str, DebianAutopkgtestResult],
    ) -> None:
        """Call self.autopkgtest.parse_summary_file(summary), assert result."""
        summary_file = self.create_temporary_file(
            contents=autopkgtest_summary.encode("utf-8")
        )
        actual = Autopkgtest._parse_summary_file(summary_file)

        self.assertEqual(actual, expected)

    def test_parse_summary_file_empty(self) -> None:
        """Parse an empty file."""
        contents = ""
        expected: dict[str, DebianAutopkgtestResult] = {}

        self.assert_parse_summary_file_expected(contents, expected)

    def test_parse_summary_file_two_tests(self) -> None:
        """Parse file with two tests and two different results."""
        contents = textwrap.dedent(
            """\
            unit-tests-server    PASS
            integration-tests-task-sbuild FLAKY
        """
        )
        expected = {
            "unit-tests-server": DebianAutopkgtestResult(status=rs.PASS),
            "integration-tests-task-sbuild": DebianAutopkgtestResult(
                status=rs.FLAKY
            ),
        }

        self.assert_parse_summary_file_expected(contents, expected)

    def test_parse_summary_file_badpkg(self) -> None:
        """Parse file with badpkg report."""
        contents = textwrap.dedent(
            """\
                blame: arg:/tmp/foo.deb deb:foo /tmp/foo.dsc
                badpkg: Test dependencies are unsatisfiable.
                erroneous package: Test dependencies are unsatisfiable.
                testbed failure: unexpected eof from the testbed
                quitting: unexpected error, see log
            """
        )
        expected: dict[str, DebianAutopkgtestResult] = {}

        with self.assertLogs("debusine.tasks.autopkgtest", level="INFO") as cm:
            self.assert_parse_summary_file_expected(contents, expected)
        self.assertEqual(
            cm.output,
            [
                "INFO:debusine.tasks.autopkgtest:Autopkgtest error: blame: "
                "arg:/tmp/foo.deb deb:foo /tmp/foo.dsc",
                "INFO:debusine.tasks.autopkgtest:Autopkgtest error: badpkg: "
                "Test dependencies are unsatisfiable.",
                "INFO:debusine.tasks.autopkgtest:Autopkgtest error: erroneous "
                "package: Test dependencies are unsatisfiable.",
                "INFO:debusine.tasks.autopkgtest:Autopkgtest error: testbed "
                "failure: unexpected eof from the testbed",
                "INFO:debusine.tasks.autopkgtest:Autopkgtest error: quitting: "
                "unexpected error, see log",
            ],
        )

    def test_parse_summary_file_expected_results(self) -> None:
        """Parse file with valid results."""
        contents = textwrap.dedent(
            """\
            test-name-1    PASS
            test-name-2     FAIL
            test-name-3     SKIP
            test-name-4     FLAKY
            *               SKIP
            """
        )
        expected = {
            "test-name-1": DebianAutopkgtestResult(status=rs.PASS),
            "test-name-2": DebianAutopkgtestResult(status=rs.FAIL),
            "test-name-3": DebianAutopkgtestResult(status=rs.SKIP),
            "test-name-4": DebianAutopkgtestResult(status=rs.FLAKY),
            "*": DebianAutopkgtestResult(status=rs.SKIP),
        }
        self.assert_parse_summary_file_expected(contents, expected)

    def test_parse_summary_file_unexpected_result_raise_value_error(
        self,
    ) -> None:
        """Parse file with unexpected result: raise ValueError."""
        contents = "test-name-1: something"
        with self.assertRaisesRegex(
            ValueError, f"Line with unexpected result: {contents}"
        ):
            self.assert_parse_summary_file_expected(contents, {})

    def test_parse_summary_file_one_test_with_details(self) -> None:
        """Parse a file with details."""
        contents = """autodep8-python3     PASS (superficial)"""
        expected = {
            "autodep8-python3": DebianAutopkgtestResult(
                status=rs.PASS,
                details="(superficial)",
            ),
        }

        self.assert_parse_summary_file_expected(contents, expected)

    def test_parse_summary_file_with_details_with_space(self) -> None:
        """Parse a file with details containing a space."""
        contents = textwrap.dedent(
            """\
            unit-tests-server    PASS
            unit-tests-client  FAIL (the details)
            unit-tests-worker  SKIP stderr: Paths: foo not set, '/tmp/test'
        """
        )
        expected = {
            "unit-tests-server": DebianAutopkgtestResult(status=rs.PASS),
            "unit-tests-client": DebianAutopkgtestResult(
                status=rs.FAIL,
                details="(the details)",
            ),
            "unit-tests-worker": DebianAutopkgtestResult(
                status=rs.SKIP,
                details="stderr: Paths: foo not set, '/tmp/test'",
            ),
        }

        self.assert_parse_summary_file_expected(contents, expected)

    def test_parse_summary_file_invalid_line(self) -> None:
        """Parse a file with an invalid line."""
        # Format of each line should be "test-name test-status"
        contents = "invalid"
        with self.assertRaisesRegex(ValueError, "^Failed to parse line"):
            self.assert_parse_summary_file_expected(contents, {})

    def test_configure_valid_schema(self) -> None:
        """Valid task data do not raise any exception."""
        self.configure_task()

    def test_configure_missing_required_source_artifact(self) -> None:
        """Missing required field "source_artifact": raise exception."""
        task_no_source_artifact = copy.deepcopy(self.SAMPLE_TASK_DATA)
        del cast(dict[str, str], task_no_source_artifact["input"])[
            "source_artifact"
        ]

        with self.assertRaises(TaskConfigError):
            self.configure_task(task_data=task_no_source_artifact)

    def test_configure_missing_required_binary_artifacts(self) -> None:
        """Missing required field "binary_artifacts": raise exception."""
        task_no_binary_artifacts = copy.deepcopy(self.SAMPLE_TASK_DATA)
        del cast(dict[str, str], task_no_binary_artifacts["input"])[
            "binary_artifacts"
        ]

        with self.assertRaises(TaskConfigError):
            self.configure_task(task_data=task_no_binary_artifacts)

    def test_configure_accept_context_artifacts(self) -> None:
        """Optional field "context_artifacts": do not raise exception."""
        task_context_artifacts_ids = copy.deepcopy(self.SAMPLE_TASK_DATA)
        cast(dict[str, list[int]], task_context_artifacts_ids["input"])[
            "context_artifacts"
        ] = [20, 21]

        self.configure_task(task_data=task_context_artifacts_ids)

    def test_configure_input_additional_property(self) -> None:
        """Additional property in input: raise exception."""
        additional_property = copy.deepcopy(self.SAMPLE_TASK_DATA)
        cast(dict[str, str], additional_property["input"])["something"] = "test"

        with self.assertRaises(TaskConfigError):
            self.configure_task(task_data=additional_property)

    def test_configure_additional_property(self) -> None:
        """Additional property: raise exception."""
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"something": "test"})

    def test_configure_environment_is_required(self) -> None:
        """Missing required field "environment": raise exception."""
        with self.assertRaises(TaskConfigError):
            self.configure_task(remove=["environment"])

    def test_configure_backend_choices(self) -> None:
        """Field "backend" accept correct values."""
        for backend in [
            BackendType.AUTO,
            BackendType.INCUS_LXC,
            BackendType.INCUS_VM,
            BackendType.UNSHARE,
        ]:
            self.configure_task(override={"backend": backend})

    def test_configure_backend_invalid_choice(self) -> None:
        """Field "backend" do not accept unexpected value."""
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"backend": "invalid"})

    def test_configure_backend_default_value(self) -> None:
        """Field "backend" default value is "auto"."""
        self.configure_task(remove=["backend"])
        self.assertEqual(self.task.data.backend, BackendType.AUTO)

    def test_configure_backend_include_tests(self) -> None:
        """Field "include_tests" is allowed."""
        self.configure_task(override={"include_tests": ["a", "b"]})

    def test_configure_backend_exclude_tests(self) -> None:
        """Field "exclude_tests" is allowed."""
        self.configure_task(override={"exclude_tests": ["a", "b"]})

    def test_configure_debug_level(self) -> None:
        """Field "debug_level" is allowed."""
        self.configure_task(override={"debug_level": 0})
        self.configure_task(override={"debug_level": 3})

    def test_configure_debug_level_too_small_too_big(self) -> None:
        """Field "debug_level" must be between 0 and 3."""
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"debug_level": -1})

        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"debug_level": 4})

    def test_configure_debug_level_default_to_0(self) -> None:
        """Field "debug_level" default value is 0."""
        self.configure_task()
        self.assertEqual(self.task.data.debug_level, 0)

    def test_configure_extra_apt_sources(self) -> None:
        """Field "extra_repositories" is allowed."""
        self.configure_task(
            override={
                "extra_repositories": [
                    {
                        "url": "http://example.net",
                        "suite": "bookworm",
                        "components": ["main"],
                    }
                ]
            }
        )

    def test_configure_use_packages_from_base_repository(self) -> None:
        """Field "use_package_from_base_repository" is allowed."""
        self.configure_task(
            override={"use_packages_from_base_repository": True}
        )

    def test_configure_use_packages_from_base_repository_default_is_false(
        self,
    ) -> None:
        """Field "use_package_from_base_repository" default value is False."""
        self.configure_task()
        self.assertFalse(self.task.data.use_packages_from_base_repository)

    def test_configure_needs_internet_accepted_values(self) -> None:
        """Field "needs_internet" accept valid values."""
        for needs_internet in ["run", "try", "skip"]:
            self.configure_task(override={"needs_internet": needs_internet})

    def test_configure_needs_internet_reject_unexpected_value(self) -> None:
        """Field "needs_internet" do not accept unexpected value."""
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"needs_internet": "yes"})

    def test_configure_needs_internet_default_value(self) -> None:
        """Field "needs_internet" default value is "run"."""
        self.configure_task()
        self.assertEqual(self.task.data.needs_internet, "run")

    def test_configure_extra_environment(self) -> None:
        """Field "extra_environment" is accepted."""
        self.configure_task(
            override={"extra_environment": {"var1": "foo1", "var2": "foo2"}}
        )

    def test_configure_fail_on_defaults(self) -> None:
        """Field "fail_on" default values are the expected ones."""
        self.configure_task()
        self.assertEqual(
            dict(self.task.data.fail_on),
            {"failed_test": True, "flaky_test": False, "skipped_test": False},
        )

    def test_configure_fail_on_failed_skipped_test_defaults(self) -> None:
        """Field "fail_on" use defaults when specifying only "flaky_test"."""
        self.configure_task(override={"fail_on": {"flaky_test": False}})

        self.assertEqual(
            dict(self.task.data.fail_on),
            {"failed_test": True, "flaky_test": False, "skipped_test": False},
        )

    def test_configure_flaky_skipped_test_defaults(self) -> None:
        """Field "fail_on" use defaults when specifying only "failed_test"."""
        self.configure_task(override={"fail_on": {"failed_test": False}})

        self.assertEqual(
            dict(self.task.data.fail_on),
            {"failed_test": False, "flaky_test": False, "skipped_test": False},
        )

    def test_configure_fail_on_unexpected(self) -> None:
        """Field "fail_on" with invalid key: raise TaskConfigError."""
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"fail_on": {"non-valid": False}})

    def test_configure_timeout_valid_values(self) -> None:
        """Field "timeout" accept the expected keys."""
        for timeout_key in [
            "global",
            "factor",
            "short",
            "install",
            "test",
            "copy",
        ]:
            self.configure_task(override={"timeout": {timeout_key: 600}})

    def test_configure_timeout_non_valid_values(self) -> None:
        """Field "timeout" invalid values must be > 0: raise TaskConfigError."""
        for timeout_key, value in [
            ("global", -1),
            ("factor", -10),
            ("short", -10),
            ("install", -20),
            ("test", -30),
            ("copy", -10),
        ]:
            with self.assertRaises(TaskConfigError):
                self.configure_task(override={"timeout": {timeout_key: value}})

    def test_configure_timeout_non_valid_key(self) -> None:
        """Field "timeout" invalid key: raise TaskConfigError."""
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"timeout": {"non-valid-key": 600}})

    def test_configure_timeout_invalid_value(self) -> None:
        """Field "timeout" do not accept unexpected value."""
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"timeout": "something"})

    def test_get_provided_worker_tags(self) -> None:
        self.mock_is_command_available({"autopkgtest": True})
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "worker:autopkgtest:version:1",
                wtags.CAP_PREFIX + "autopkgtest",
            ],
        )

    def test_get_provided_worker_tags_autopkgtest_not_available(self) -> None:
        self.mock_is_command_available({"autopkgtest": False})
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [wtags.TASK_PREFIX + "worker:autopkgtest:version:1"],
        )

    def test_analyze_worker(self) -> None:
        """Test the analyze_worker() method."""
        self.mock_is_command_available({"autopkgtest": True})
        metadata = self.task.analyze_worker()
        self.assertEqual(metadata["autopkgtest:available"], True)

    def test_analyze_worker_autopkgtest_not_available(self) -> None:
        """analyze_worker() handles autopkgtest not being available."""
        self.mock_is_command_available({"autopkgtest": False})
        metadata = self.task.analyze_worker()
        self.assertEqual(metadata["autopkgtest:available"], False)

    def test_can_run_on(self) -> None:
        """can_run_on returns True if unshare and autopkgtest are available."""
        architecture = self.task.data.build_architecture
        self.assertTrue(
            self.can_run_on(
                {
                    "system:worker_type": WorkerType.EXTERNAL,
                    "system:architectures": [architecture],
                    "executor:unshare:available": True,
                    "autopkgtest:available": True,
                    "autopkgtest:version": self.task.TASK_VERSION,
                }
            )
        )

    def test_can_run_on_mismatched_task_version(self) -> None:
        """can_run_on returns False for mismatched task versions."""
        self.assertFalse(
            self.can_run_on(
                {
                    "system:worker_type": WorkerType.EXTERNAL,
                    "executor:unshare:available": True,
                    "autopkgtest:available": True,
                    "autopkgtest:version": self.task.TASK_VERSION + 1,
                }
            )
        )

    def test_can_run_on_unshare_not_available(self) -> None:
        """can_run_on returns False if unshare is not available."""
        self.assertFalse(
            self.can_run_on(
                {
                    "system:worker_type": WorkerType.EXTERNAL,
                    "executor:unshare:available": False,
                    "autopkgtest:available": True,
                    "autopkgtest:version": self.task.TASK_VERSION,
                }
            )
        )

    def test_can_run_on_autopkgtest_not_available(self) -> None:
        """can_run_on returns False if autopkgtest is not available."""
        self.assertFalse(
            self.can_run_on(
                {
                    "system:worker_type": WorkerType.EXTERNAL,
                    "executor:unshare:available": True,
                    "autopkgtest:available": False,
                    "autopkgtest:version": self.task.TASK_VERSION,
                }
            )
        )

    def create_binary_package_data(
        self, srcpkg_name: str
    ) -> DebianBinaryPackage:
        """Create sample `debian:binary-package` artifact data."""
        return DebianBinaryPackage(
            srcpkg_name=srcpkg_name,
            srcpkg_version="1.0",
            deb_fields={
                "Package": srcpkg_name,
                "Version": "1.0",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        )

    def create_upload_data(self, name: str) -> DebianUpload:
        """Create sample `debian:upload` artifact data."""
        return DebianUpload(
            type="dpkg",
            changes_fields={
                "Architecture": "amd64",
                "Files": [{"name": f"{name}_1.0_amd64.deb"}],
            },
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            provide_system={ttags.SOURCE_PACKAGE_PREFIX + "hello"},
            require_system={
                wtags.BUILD_ARCH_PREFIX + "amd64",
                wtags.EXECUTOR_PREFIX + "unshare",
                wtags.CAP_PREFIX + "autopkgtest",
            },
        )

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        assert self.compute_dynamic_data(self.task) == AutopkgtestDynamicData(
            environment_id=1,
            input_source_artifact_id=5,
            input_binary_artifacts_ids=[10, 11],
            input_context_artifacts_ids=[],
            subject="hello",
            parameter_summary="hello_1.0 (amd64)",
            runtime_context="amd64:unshare",
            configuration_context="sid",
        )

    def test_get_subject(self) -> None:
        self.assertEqual(self.get_subject(), "hello")

    def test_get_configuration_context(self) -> None:
        self.assertEqual(self.get_configuration_context(), "sid")

    def test_compute_dynamic_data_context(self) -> None:
        """Dynamic data includes input context artifact IDs if relevant."""
        self.resolver.set_field_artifacts(
            "input_binary_artifacts",
            BaseTaskInputArtifact(
                artifact_id=10,
                category=ArtifactCategory.BINARY_PACKAGE,
                data=self.create_binary_package_data("hello"),
            ),
            BaseTaskInputArtifact(
                artifact_id=12,
                category=ArtifactCategory.UPLOAD,
                data=self.create_upload_data("hello"),
            ),
        )
        self.resolver.set_field_artifacts(
            "input_context_artifacts",
            BaseTaskInputArtifact(
                artifact_id=2,
                category=ArtifactCategory.BINARY_PACKAGE,
                data=self.create_binary_package_data("context"),
            ),
            BaseTaskInputArtifact(
                artifact_id=4,
                category=ArtifactCategory.UPLOAD,
                data=self.create_upload_data("context"),
            ),
        )
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            AutopkgtestDynamicData(
                environment_id=1,
                input_source_artifact_id=5,
                input_binary_artifacts_ids=[10, 12],
                input_context_artifacts_ids=[2, 4],
                subject="hello",
                parameter_summary="hello_1.0 (amd64)",
                runtime_context="amd64:unshare",
                configuration_context="sid",
            ),
        )

    def test_acceptable_input_categories(self) -> None:
        self.assertEqual(
            Autopkgtest.inputs["input_source_artifact"].categories,  # type: ignore[attr-defined]
            (ArtifactCategory.SOURCE_PACKAGE, ArtifactCategory.UPLOAD),
        )
        self.assertEqual(
            Autopkgtest.inputs["input_binary_artifacts"].categories,  # type: ignore[attr-defined]
            (ArtifactCategory.BINARY_PACKAGE, ArtifactCategory.UPLOAD),
        )
        self.assertEqual(
            Autopkgtest.inputs["input_context_artifacts"].categories,  # type: ignore[attr-defined]
            (ArtifactCategory.BINARY_PACKAGE, ArtifactCategory.UPLOAD),
        )

    def test_get_input_artifacts_ids(self) -> None:
        """Test get_input_artifacts_ids."""
        self.task.dynamic_data = None
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

        self.task.dynamic_data = AutopkgtestDynamicData(
            environment_id=1,
            input_source_artifact_id=2,
            input_binary_artifacts_ids=[3],
            input_context_artifacts_ids=[4, 5],
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [1, 2, 3, 4, 5])

    def test_cmdline_executor_not_set(self) -> None:
        """Raise AssertionError: self.executor is not set."""
        msg = (
            r"^self\.executor not set - self\._prepare_for_execution\(\) "
            r"must be called before _cmdline\(\)$"
        )
        self.assertRaisesRegex(AssertionError, msg, self.task._cmdline)

    def test_cmdline_minimum_options(self) -> None:
        """Command line has the minimum options."""
        self.mock_executor()
        minimum_options = [
            "autopkgtest",
            "--apt-upgrade",
            f"--output-dir={Autopkgtest.ARTIFACT_DIR}",
            f"--summary={Autopkgtest.SUMMARY_FILE}",
            "--no-built-binaries",
        ]
        cmdline = self.task._cmdline()
        self.assertEqual(cmdline[: len(minimum_options)], minimum_options)

    def test_cmdline_include_tests(self) -> None:
        """Command line include --test-name=XXX."""
        self.mock_executor()
        include_tests = ["test1", "test2"]

        self.task.data.include_tests = include_tests

        cmdline = self.task._cmdline()

        for include_test in include_tests:
            self.assertIn(f"--test-name={include_test}", cmdline)

    def test_cmdline_exclude_tests(self) -> None:
        """Command line include --skip-test=XXX."""
        self.mock_executor()
        exclude_tests = ["test1", "test2"]

        self.task.data.exclude_tests = exclude_tests

        cmdline = self.task._cmdline()

        for exclude_test in exclude_tests:
            self.assertIn(f"--skip-test={exclude_test}", cmdline)

    def test_cmdline_debug_level(self) -> None:
        """Command line include "-dd" (debug_level)."""
        self.mock_executor()
        self.task.data.debug_level = 2

        cmdline = self.task._cmdline()

        self.assertIn("-dd", cmdline)

    def test_cmdline_extra_repositories_trixie(self) -> None:
        """Ensure the --copy= parameter is passed to autopkgtest."""
        self.set_dynamic_data()
        debusine_mock = self.mock_debusine()
        self.mock_image_download(debusine_mock)
        self.task._prepare_executor()
        self.task.extra_repository_sources = [
            Path("/tmp/extra_repository_0.sources")
        ]
        self.task.extra_repository_keys = [
            Path("/tmp/key.asc"),
        ]
        cmdline = self.task._cmdline()

        self.assertIn(
            (
                "--copy=/tmp/extra_repository_0.sources:"
                "/etc/apt/sources.list.d/extra_repository_0.sources"
            ),
            cmdline,
        )
        self.assertIn("--copy=/tmp/key.asc:/etc/apt/keyrings/key.asc", cmdline)

    def test_cmdline_extra_debusine_repositories(self) -> None:
        """Extra repositories can be hosted by Debusine."""
        signing_key = "\n".join(
            (
                "-----BEGIN PGP PUBLIC KEY BLOCK-----",
                "",
                "ABCDEFGHI",
                "-----END PGP PUBLIC KEY BLOCK-----",
            )
        )
        self.resolver.field_values["extra_repositories"] = [
            ExtraExternalRepository(
                url="http://deb.example.com/scope/workspace",
                suite="trixie-local",
                components=["main"],
                signing_key=signing_key,
            )
        ]
        self.set_dynamic_data()

        debusine_mock = self.mock_debusine()
        self.mock_image_download(debusine_mock)
        self.task._prepare_executor()
        destination = self.create_temporary_directory()

        self.task.write_extra_repository_config("trixie", destination)
        cmdline = self.task._cmdline()

        self.assertEqual(self.task.extra_repository_keys, [])
        [extra_repository_source] = self.task.extra_repository_sources
        self.assertEqual(
            extra_repository_source.read_text(),
            textwrap.dedent(
                """\
                Types: deb
                URIs: http://deb.example.com/scope/workspace
                Suites: trixie-local
                Components: main
                Signed-By:
                 -----BEGIN PGP PUBLIC KEY BLOCK-----
                 .
                 ABCDEFGHI
                 -----END PGP PUBLIC KEY BLOCK-----
                """
            ),
        )
        self.assertIn(
            f"--copy={extra_repository_source}:"
            f"/etc/apt/sources.list.d/{extra_repository_source.name}",
            cmdline,
        )

    def test_cmdline_apt_default_release_not_included(self) -> None:
        """Command line NOT include "--apt-default-release={distribution}."""
        self.mock_executor()
        self.task.data.use_packages_from_base_repository = False
        cmdline = self.task._cmdline()
        self.assertNotIn("--apt-default-release=bookworm", cmdline)

    def test_cmdline_apt_default_release_included(self) -> None:
        """Command line include --"apt-default-release=distribution"."""
        self.mock_executor()
        self.task.data.use_packages_from_base_repository = True
        cmdline = self.task._cmdline()
        self.assertIn("--apt-default-release=bookworm", cmdline)

    def test_cmdline_pin_packages_not_included(self) -> None:
        self.mock_executor()
        self.task.data.pin_packages = {}
        cmdline = self.task._cmdline()
        self.assertNotIn(
            "--pin-packages", [arg.split("=")[0] for arg in cmdline]
        )

    def test_cmdline_pin_packages_included(self) -> None:
        self.mock_executor()
        self.task.data.pin_packages = {
            "unstable": ["src:glibc", "hello"],
            "experimental": ["src:python3-defaults"],
        }
        cmdline = self.task._cmdline()
        self.assertIn("--pin-packages=unstable=src:glibc,hello", cmdline)
        self.assertIn(
            "--pin-packages=experimental=src:python3-defaults", cmdline
        )

    def test_cmdline_pin_packages_empty_pkglist(self) -> None:
        self.mock_executor()
        self.task.data.pin_packages = {"unstable": []}
        cmdline = self.task._cmdline()
        self.assertIn("--pin-packages=unstable=", cmdline)

    def test_cmdline_environment(self) -> None:
        """Command line include "--env=VAR=VALUE" for environment variables."""
        self.mock_executor()
        extra_environment = {"DEBUG": "1", "LOG": "NONE"}
        self.task.data.extra_environment = extra_environment

        cmdline = self.task._cmdline()

        for var, value in extra_environment.items():
            self.assertIn(f"--env={var}={value}", cmdline)

    def test_cmdline_includes_executor(self) -> None:
        """Command line include "-- backend"."""
        self.set_dynamic_data()
        debusine_mock = self.mock_debusine()
        self.mock_image_download(debusine_mock)

        self.task._prepare_executor()

        cmdline = self.task._cmdline()

        backend = "unshare"
        backend_index = cmdline.index(backend)
        self.assertGreater(backend_index, 0)
        self.assertEqual(
            cmdline[backend_index + 1 :],
            [
                "--arch",
                "amd64",
                "--release",
                "bookworm",
                "--tarball",
                str(self.image_cache_path / "42/system.tar.xz"),
            ],
        )

    def test_cmdline_add_targets(self) -> None:
        """Targets for autopkgtest are added."""
        self.mock_executor()
        binary_path = Path("/tmp/download-dir/package.deb")
        source_path = Path("/tmp/download-dir/package.dsc")

        self.task._autopkgtest_targets = [binary_path, source_path]
        cmdline = self.task._cmdline()

        binary_path_index = cmdline.index(str(binary_path))

        self.assertGreater(binary_path_index, 0)
        self.assertEqual(cmdline[binary_path_index + 1], str(source_path))

    def test_needs_internet(self) -> None:
        """Command line include "--needs-internet"."""
        self.mock_executor()
        needs_internet = AutopkgtestNeedsInternet.TRY

        self.task.data.needs_internet = needs_internet

        cmdline = self.task._cmdline()

        self.assertIn(f"--needs-internet={needs_internet}", cmdline)

    def test_timeout(self) -> None:
        """Command line include "--timeout-KEY=VALUE."."""
        self.mock_executor()
        timeout = {"global": 60, "factor": 600}

        self.task.data.timeout = AutopkgtestTimeout(**timeout)

        cmdline = self.task._cmdline()

        for key, value in timeout.items():
            self.assertIn(f"--timeout-{key}={value}", cmdline)
        for key in ("global_", "short", "install", "test", "copy", "copy_"):
            self.assertNotIn(
                f"--timeout-{key}", [arg.split("=")[0] for arg in cmdline]
            )

    def test_cmdline_unshare_copy_resolv_conf(self) -> None:
        """In unshare mode, we copy /etc/resolv.conf if it exists."""
        self.mock_executor()
        self.task.data.backend = BackendType.UNSHARE

        with patch("pathlib.Path.is_file", return_value=True):
            cmdline = self.task._cmdline()

        self.assertIn("--copy=/etc/resolv.conf:/etc/resolv.conf", cmdline)

    def test_cmdline_unshare_missing_resolv_conf(self) -> None:
        """In unshare mode, we don't copy /etc/resolv.conf if it's missing."""
        self.mock_executor()
        self.task.data.backend = BackendType.UNSHARE

        with patch("pathlib.Path.is_file", return_value=False):
            cmdline = self.task._cmdline()

        self.assertNotIn("--copy=/etc/resolv.conf:/etc/resolv.conf", cmdline)

    def test_cmdline_not_unshare_does_not_copy_resolv_conf(self) -> None:
        """For non-unshare backends, we don't copy /etc/resolv.conf."""
        self.mock_executor()
        self.task.data.backend = BackendType.INCUS_LXC

        with patch("pathlib.Path.is_file", return_value=True):
            cmdline = self.task._cmdline()

        self.assertNotIn("--copy=/etc/resolv.conf:/etc/resolv.conf", cmdline)

    def test_select_default_backend(self) -> None:
        """We select Autopkgtest.DEFAULT_BACKEND if backend is auto."""
        self.set_dynamic_data()
        debusine_mock = self.mock_debusine()
        self.mock_image_download(debusine_mock)

        self.task.data.backend = BackendType.AUTO

        self.task._prepare_executor()

        assert self.task.executor

        self.assertEqual(
            self.task.executor.backend_name,
            Autopkgtest.DEFAULT_BACKEND,
        )

    def test_select_specified_backend(self) -> None:
        """We select the specified backend if backend is not auto."""
        self.set_dynamic_data()
        debusine_mock = self.mock_debusine()
        self.mock_image_download(debusine_mock)

        # FIXME: Don't use the default when we support more backends.
        self.task.data.backend = BackendType.UNSHARE

        self.task._prepare_executor()

        assert self.task.executor

        self.assertEqual(self.task.executor.backend_name, BackendType.UNSHARE)

    def test_environment_codename_executor(self) -> None:
        """Test _environment_codename gets the executor distribution."""
        self.mock_executor()
        self.assertEqual(self.task._environment_codename(), "bookworm")

    def test_environment_codename_executor_unconfigured(self) -> None:
        """Test _environment_codename raises ValueError without executor."""
        with self.assertRaisesRegex(
            ValueError,
            r"self\.executor not set, call _prepare_executor\(\) first",
        ):
            self.task._environment_codename()

    def test_fetch_input_no_context_artifacts(self) -> None:
        """Download "source_artifact" and "binary_artifacts"."""
        self.set_dynamic_data()
        debusine_mock = self.mock_debusine()
        self.mock_image_download(debusine_mock)
        workspace = "Testing"
        debusine_mock.download_artifact.return_value = create_artifact_response(
            id=10,
            workspace=workspace,
            category="Test",
            created_at=dt.datetime.now(tz=dt.UTC),
            data={},
            download_tar_gz_url="https://example.com/not-used",
            files=FilesResponseType({}),
            files_to_upload=[],
        )

        workspace = "Testing"
        source_package_url = "http://example.com/artifact/15/hello.dsc"

        files = {
            "hello.orig.tar.gz": FileResponse(
                size=4000,
                checksums={},
                type="file",
                url="http://example.com/artifact/1/hello.orig.tar.gz",
            ),
            "hello.dsc": FileResponse(
                size=258, checksums={}, type="file", url=source_package_url
            ),
        }

        data_input = self.task.data.input

        debusine_mock.download_artifact.return_value = create_artifact_response(
            id=10,
            workspace=workspace,
            category="Test",
            created_at=dt.datetime.now(tz=dt.UTC),
            data={},
            download_tar_gz_url="https://example.com/not-used",
            files=FilesResponseType(files),
            files_to_upload=[],
        )

        destination = self.create_temporary_directory()

        self.task.fetch_input(destination)

        calls = [
            call(
                data_input.source_artifact,
                destination,
                tarball=False,
                workspace="testing",
            )
        ]

        calls += [
            call(
                binary_artifact_id,
                destination,
                tarball=False,
                workspace="testing",
            )
            for binary_artifact_id in data_input.binary_artifacts
        ]

        debusine_mock.download_artifact.assert_has_calls(calls)

        self.assertEqual(str(self.task._source_package_url), source_package_url)
        self.assertEqual(self.task._source_package_path, "hello.dsc")

    def test_fetch_input_context_artifacts(self) -> None:
        """Download the "context_artifacts"."""
        self.set_dynamic_data()
        debusine_mock = self.mock_debusine()
        self.mock_image_download(debusine_mock)

        destination = self.create_temporary_directory()

        data_input = self.task.data.input
        data_input.context_artifacts = LookupMultiple((50, 51))
        assert self.task.dynamic_data
        self.task.dynamic_data.input_context_artifacts_ids = [50, 51]

        self.task.fetch_input(destination)

        calls = [
            call(
                context_artifact_id,
                destination,
                tarball=False,
                workspace="testing",
            )
            for context_artifact_id in data_input.context_artifacts
        ]

        debusine_mock.download_artifact.assert_has_calls(calls)

    def test_configure_for_execution(self) -> None:
        """Autopkgtest configure_for_execution() return True."""
        self.mock_executor()

        directory = self.create_temporary_directory()

        self.task._source_package_path = "hello.dsc"

        dsc = self.write_dsc_example_file(
            directory / self.task._source_package_path
        )

        self.assertTrue(self.task.configure_for_execution(directory))

        self.assertEqual(self.task._source_package_name, dsc["Source"])
        self.assertEqual(self.task._source_package_version, dsc["Version"])

    def test_configure_for_execution_invalid_dsc(self) -> None:
        """configure_for_execution() handles an invalid .dsc file."""
        directory = self.create_temporary_directory()
        self.task._source_package_path = "hello.dsc"
        (directory / self.task._source_package_path).touch()

        self.assertFalse(self.task.configure_for_execution(directory))

        log_file_contents = (
            self.task.log_directory / "configure_for_execution.log"
        ).read_text()
        self.assertEqual(
            log_file_contents, "'hello.dsc' does not contain Source field\n"
        )

    def setup_task_result(
        self, test_result: str, fail_on_config: dict[str, bool]
    ) -> Path:
        """
        Write autopkgtest summary file with one test output, configure task.

        :param test_result: autopkgtest test result (FAIL, FLAKY...)
        :param fail_on_config: configure task with fail_on = fail_on_config.
        :return: directory with the summary file.
        """
        directory = self.create_temporary_directory()
        (directory / Autopkgtest.ARTIFACT_DIR).mkdir()
        summary_file = directory / Autopkgtest.SUMMARY_FILE
        summary_file.write_text(test_result)

        self.configure_task(override={"fail_on": fail_on_config})

        return directory

    def test_task_result_failed_test_fail_on_failed(self) -> None:
        """Test failed, fail on failed_test=True. Failure."""
        directory = self.setup_task_result(
            "test_name\tFAIL", {"failed_test": True}
        )
        self.assertEqual(
            self.task.task_result(0, directory), WorkRequestResults.FAILURE
        )

    def test_task_result_failed_test_do_not_fail_on_failed(self) -> None:
        """Test failed, fail on failed_test=False. Success."""
        directory = self.setup_task_result(
            "test_name\tFAIL", {"failed_test": False}
        )
        self.assertEqual(
            self.task.task_result(0, directory), WorkRequestResults.SUCCESS
        )

    def test_task_result_flaky_test_fail_on_flaky_failure(self) -> None:
        """Test is flaky, fail on flaky_test=True. Failure."""
        directory = self.setup_task_result(
            "test_name\tFLAKY", {"flaky_test": True}
        )
        self.assertEqual(
            self.task.task_result(0, directory), WorkRequestResults.FAILURE
        )

    def test_task_result_flaky_test_do_not_fail_on_flaky_success(self) -> None:
        """Test is flaky, fail on flaky_test=False. Success."""
        directory = self.setup_task_result(
            "test_name\tFLAKY", {"flaky_test": False}
        )
        self.assertEqual(
            self.task.task_result(0, directory), WorkRequestResults.SUCCESS
        )

    def test_task_result_skipped_test_fail_on_skipped_success(self) -> None:
        """Test is skipped, fail on skipped_test=True. Failure."""
        directory = self.setup_task_result(
            "test_name\tSKIP", {"skipped_test": True}
        )
        self.assertEqual(
            self.task.task_result(8, directory), WorkRequestResults.FAILURE
        )

    def test_task_result_all_skipped(self) -> None:
        """Tests are skipped. Skipped."""
        directory = self.setup_task_result(
            "test_name\tSKIP", {"skipped_test": False}
        )
        self.assertEqual(
            self.task.task_result(8, directory), WorkRequestResults.SKIPPED
        )

    def test_task_result_skipped_test_do_not_fail_on_skipped_failure(
        self,
    ) -> None:
        """A test is skipped, fail on skipped_test=False. Success."""
        directory = self.setup_task_result(
            "test_name\tSKIP\ntest_2\tPASS", {"skipped_test": False}
        )
        self.assertEqual(
            self.task.task_result(0, directory), WorkRequestResults.SUCCESS
        )

    def test_task_result_ignore_return_code(self) -> None:
        """Summary file is inconclusive and ignore return value."""
        directory = self.setup_task_result(
            "foo: nothing", {"failed_test": False}
        )
        self.assertEqual(
            self.task.task_result(4, directory), WorkRequestResults.SUCCESS
        )

    def test_task_result_flaky_return_code(self) -> None:
        """Summary file is inconclusive and return value 2."""
        directory = self.setup_task_result("foo: nothing", {"flaky_test": True})
        self.assertEqual(
            self.task.task_result(2, directory), WorkRequestResults.FAILURE
        )

    def test_task_result_skip_return_code(self) -> None:
        """Summary file is inconclusive and return value 8."""
        directory = self.setup_task_result(
            "foo: nothing", {"skipped_test": True}
        )
        self.assertEqual(
            self.task.task_result(8, directory), WorkRequestResults.FAILURE
        )

    def test_task_result_fail_on_return_code(self) -> None:
        """Summary file is inconclusive fail on return value."""
        directory = self.setup_task_result(
            "erroneous package: Test dependencies are unsatisfiable.",
            {"failed_test": True},
        )
        self.assertEqual(
            self.task.task_result(12, directory), WorkRequestResults.FAILURE
        )

    def test_upload_artifacts(self) -> None:
        """Upload the AutopkgtestArtifact."""
        exec_dir = self.create_temporary_directory()

        # Debusine.upload_artifact is mocked to verify the call only
        debusine_mock = self.mock_debusine()
        self.mock_executor()

        # Worker would set the server
        self.task.configure_server_access(debusine_mock)

        results = {
            "testname": DebianAutopkgtestResult(status=rs.PASS),
        }
        self.task._parsed = results

        self.task._source_package_path = "test"
        self.task._source_package_name = "hello"
        self.task._source_package_version = "2.0.5b"
        self.task._source_package_url = pydantic.AnyUrl(
            "https://some-url.com/test"
        )

        source_artifact_id = 5
        binary_artifact_ids = [10, 11]
        self.task.dynamic_data = AutopkgtestDynamicData(
            environment_id=1,
            input_source_artifact_id=source_artifact_id,
            input_binary_artifacts_ids=binary_artifact_ids,
            input_context_artifacts_ids=[],
        )

        uploaded_artifact = create_remote_artifact(
            id=42, workspace=self.task.workspace_name
        )

        debusine_mock.upload_artifact.return_value = uploaded_artifact

        self.task.upload_artifacts(
            exec_dir, execution_result=WorkRequestResults.SUCCESS
        )

        expected_autopkgtest_artifact = AutopkgtestArtifact.create(
            exec_dir,
            DebianAutopkgtest(
                results=results,
                cmdline=shlex.join(self.task._cmdline()),
                source_package=DebianAutopkgtestSource(
                    name="hello",
                    version="2.0.5b",
                    url="https://some-url.com/test",
                ),
                architecture=self.task.data.build_architecture,
                distribution="debian:bookworm",
            ),
        )

        debusine_mock.upload_artifact.assert_called_once_with(
            expected_autopkgtest_artifact,
            workspace=self.task.workspace_name,
            work_request=self.task.work_request_id,
        )

        self.assertEqual(debusine_mock.relation_create.call_count, 3)
        debusine_mock.relation_create.assert_has_calls(
            [
                call(
                    uploaded_artifact.id,
                    source_artifact_id,
                    RelationType.RELATES_TO,
                ),
                call(uploaded_artifact.id, 10, RelationType.RELATES_TO),
                call(uploaded_artifact.id, 11, RelationType.RELATES_TO),
            ]
        )

    def test_upload_artifacts_without_configure_server_access(self) -> None:
        """upload_artifacts() asserts that self.debusine is set."""
        self.assertRaisesRegex(
            AssertionError,
            r"^self\.debusine not set",
            self.task.upload_artifacts,
            self.create_temporary_directory(),
            execution_result=WorkRequestResults.SUCCESS,
        )

    def test_upload_artifacts_without_executor(self) -> None:
        """upload_artifacts() asserts that self.executor is set."""
        self.task.debusine = MagicMock(spec=Debusine)

        msg = (
            r"^self\.executor not set - self\._prepare_for_execution\(\) "
            r"must be called before upload_artifacts\(\)$"
        )
        self.assertRaisesRegex(
            AssertionError,
            msg,
            self.task.upload_artifacts,
            self.create_temporary_directory(),
            execution_result=WorkRequestResults.SUCCESS,
        )

    def test_check_directory_for_consistency_missing_summary(self) -> None:
        """Return missing ARTIFACT_DIR/summary file."""
        build_directory = self.create_temporary_directory()

        self.assertEqual(
            self.task.check_directory_for_consistency_errors(build_directory),
            [f"'{self.task.SUMMARY_FILE}' does not exist"],
        )

    def test_check_directory_for_consistency_all_good(self) -> None:
        """Return no errors ARTIFACT_DIR/summary file."""
        build_directory = self.create_temporary_directory()
        (build_directory / self.task.ARTIFACT_DIR).mkdir()
        (build_directory / self.task.ARTIFACT_DIR / "summary").write_text(
            "test-results"
        )

        self.assertEqual(
            self.task.check_directory_for_consistency_errors(build_directory),
            [],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command manage_worker."""

import datetime as dt
from typing import ClassVar, override

from django.utils import timezone

from debusine.db.models import Token, WorkRequest, Worker
from debusine.django.management.tests import call_command
from debusine.test.django import TestCase


class ManageWorkerCommandTests(TestCase):
    """Tests for manage_worker command."""

    token: ClassVar[Token]
    worker: ClassVar[Worker]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Set up a default token and worker."""
        super().setUpTestData()
        cls.token = cls.playground.create_worker_token(
            enabled=False, set_worker=False
        )
        cls.worker = Worker.objects.create(
            name='worker-a',
            token=cls.token,
            registered_at=timezone.now(),
        )
        cls.token.refresh_from_db()

    def test_enable_forwards_with_deprecation_warning(self) -> None:
        """`manage_worker enable` forwards to `worker enable` and warns."""
        self.assertFalse(self.token.enabled)

        with self.assertWarnsMessage(
            DeprecationWarning,
            "The `debusine-admin manage_worker enable` command has been "
            "deprecated in favour of `debusine-admin worker enable`",
        ):
            call_command("manage_worker", "enable", self.worker.name)

        self.token.refresh_from_db()
        self.assertTrue(self.token.enabled)

    def test_disable_forwards_with_deprecation_warning(self) -> None:
        """`manage_worker disable` forwards to `worker disable` and warns."""
        self.token.enable()
        self.token.refresh_from_db()

        work_request_finished = self.playground.create_worker_task(
            worker=self.worker,
            started_at=timezone.now(),
            result=WorkRequest.Results.SUCCESS,
        )
        work_request_running = self.playground.create_worker_task(
            worker=self.worker, mark_running=True, started_at=timezone.now()
        )
        work_request_pending = self.playground.create_worker_task(
            mark_pending=True, started_at=timezone.now()
        )
        work_request_pending.worker = self.worker
        work_request_pending.save()

        self.assertTrue(self.token.enabled)

        with self.assertWarnsMessage(
            DeprecationWarning,
            "The `debusine-admin manage_worker disable` command has been "
            "deprecated in favour of `debusine-admin worker disable`",
        ):
            call_command("manage_worker", "disable", self.worker.name)

        # Worker is disabled
        self.token.refresh_from_db()
        self.assertFalse(self.token.enabled)

        # work requests in running or pending status are not assigned
        # to the worker anymore
        for work_request in work_request_running, work_request_pending:
            work_request.refresh_from_db()
            self.assertIsNone(work_request.worker)
            self.assertEqual(work_request.status, WorkRequest.Statuses.PENDING)
            self.assertIsNone(work_request.started_at)

        # work request that was completed: fields did not change
        work_request_finished.refresh_from_db()
        self.assertEqual(
            work_request_finished.status, WorkRequest.Statuses.COMPLETED
        )
        self.assertEqual(work_request_finished.worker, self.worker)
        self.assertIsInstance(work_request_finished.started_at, dt.datetime)



"""Appropriate settings to run the test suite."""

from debusine.signing.settings import defaults
from debusine.signing.settings.development import *  # noqa: F401, F403

# Restore INSTALLED_APPS from defaults to disable debug_toolbar
INSTALLED_APPS = defaults.INSTALLED_APPS.copy()

# Some checks need to know if we are running tests or not
# Test run with DEBUG=False but it is allowed, for example, to use
# the default SECRET_KEY during tests
TEST_MODE = True

# When running the test suite, enable all apps so that we have all the models
INSTALLED_APPS.extend([])



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command list_users."""

from typing import override

from django.contrib.auth import get_user_model

from debusine.django.management.tests import call_command
from debusine.server.management.commands.tests.utils import TabularOutputTests
from debusine.test.django import TestCase


class ListUsersCommandTests(TabularOutputTests, TestCase):
    """Tests for the list_users command."""

    @override
    def setUp(self) -> None:
        """Set up new user for the tests."""
        super().setUp()
        self.user = get_user_model().objects.create_user(
            username="bob", email="bob@bob.com", password="123456"
        )

    def test_list_tokens_no_filtering(self) -> None:
        """list_users print the user's information."""
        with self.assertPrintsTable() as output:
            _, stderr, exit_code = call_command("list_users")

        self.assertEqual(output.col(0), [self.user.username])
        self.assertEqual(output.col(1), [self.user.email])
        self.assertEqual(output.col(2), [self.user.date_joined.isoformat()])

        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine command line interface, collection management commands."""

import abc
import argparse
import sys
from collections.abc import Iterable
from typing import cast, override

import rich
import rich.json
from rich.table import Table

from debusine.artifacts.models import CollectionCategory
from debusine.client.commands.base import (
    InputDataCommand,
    ModelCommand,
    OptionalInputDataCommand,
    RequiredInputDataCommand,
    WorkspaceCommand,
)
from debusine.client.exceptions import DebusineError
from debusine.client.models import (
    CollectionData,
    CollectionDataNew,
    LookupChildType,
    LookupMultiple,
    LookupMultipleResponse,
    LookupSingleResponse,
)


class CollectionCommand(
    ModelCommand[CollectionData], WorkspaceCommand, abc.ABC
):
    """Common infrastructure for collection commands."""

    def markup_category(self, collection: CollectionData) -> str:
        """Markup a collection category for rich."""
        if collection.category == CollectionCategory.WORKFLOW_INTERNAL:
            return collection.category
        category_url = self.debusine.webui_url(
            collection.get_link_url("webui_category")
        )
        return f"[link={category_url}]{collection.category}[/]"

    def _url(self, instance: CollectionData) -> str:
        return self.debusine.webui_url(instance.get_link_url("webui_self"))

    @override
    def _list_rich(self, collections: Iterable[CollectionData]) -> None:
        table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
        table.add_column("ID")
        table.add_column("Name")
        table.add_column("Category")
        table.add_column("Retention (history)")
        table.add_column("Retention (metadata)")
        table.add_column("Data")
        for collection in collections:
            url = self._url(collection)
            table.add_row(
                f"{collection.id}",
                f"[link={url}]{collection.name}[/]",
                self.markup_category(collection),
                (
                    str(period)
                    if (period := collection.full_history_retention_period)
                    is not None
                    else "unlimited"
                ),
                (
                    str(period)
                    if (period := collection.metadata_only_retention_period)
                    is not None
                    else "unlimited"
                ),
                (
                    rich.json.JSON.from_data(collection.data)
                    if collection.data
                    else None
                ),
            )
        rich.print(table)

    @override
    def _show_rich(self, collection: CollectionData) -> None:
        """Show a collection for humans."""
        url = self._url(collection)
        table = Table(box=rich.box.SIMPLE, show_header=False)
        table.add_column("Name", justify="right")
        table.add_row("ID:", f"#{collection.id}")
        table.add_row("Web URL:", f"[link={url}]{url}[/]")
        table.add_row("Name:", collection.name)
        table.add_row("Category:", self.markup_category(collection))
        full = (
            f"{period} days"
            if (period := collection.full_history_retention_period) is not None
            else "unlimited"
        )
        meta = (
            f"{period} days"
            if (period := collection.metadata_only_retention_period) is not None
            else "unlimited"
        )
        table.add_row("Full history retention:", full)
        table.add_row("Metadata only retention:", meta)
        table.add_row("Data:", rich.json.JSON.from_data(collection.data))
        if collection.stats:
            table.add_row("Stats:", rich.json.JSON.from_data(collection.stats))
        rich.print(table)


class List(CollectionCommand, group="collection"):
    """List collections in a workspace."""

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            collections = self.debusine.collection_iter(self.workspace)
            self.list(collections)


class SingleCollectionCommand(WorkspaceCommand, abc.ABC):
    """Command operating on a single collection."""

    def __init__(
        self, parser: argparse.ArgumentParser, args: argparse.Namespace
    ) -> None:
        """Identify the selected collection."""
        super().__init__(parser, args)
        if "@" in (cat := args.category_or_fullname):
            self.collection_name, self.collection_category = cat.split("@", 1)
            if args.name is not None:
                parser.error(
                    "collection name is given"
                    f" both in `{args.category_or_fullname}`"
                    " and in the name argument"
                )
        else:
            self.collection_category = cat
            self.collection_name = args.name or "_"

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "category_or_fullname",
            type=str,
            metavar="[name@]category",
            help="collection category or name@category full name",
        )
        parser.add_argument(
            "name",
            type=str,
            nargs="?",
            help="collection name, only needed if not specified"
            " in the first argument. Defaults to '_'.",
        )


class Show(SingleCollectionCommand, CollectionCommand, group="collection"):
    """Show an existing collection."""

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            collection = self.debusine.collection_get(
                self.workspace,
                self.collection_category,
                self.collection_name,
                with_stats=True,
            )
        self.show(collection)


def retention_period(value: str) -> int | str:
    """Validate a retention period from the command line."""
    if value.isdigit():
        return int(value)
    elif value == "-":
        return "-"
    else:
        raise argparse.ArgumentError(
            argument=None, message="retention period must be a number or '-'"
        )


class CollectionInput(CollectionCommand, abc.ABC):
    """Common code for commands that input data for collections."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "--full-history-retention-period",
            type=retention_period,
            help="set full history retention period (in days)."
            " Use a dash ('-') to unset",
        )
        parser.add_argument(
            "--metadata-only-retention-period",
            type=retention_period,
            help="set metadata only retention period (in days)."
            " Use a dash ('-') to unset",
        )

    def update_collection(
        self, collection: CollectionDataNew | CollectionData
    ) -> None:
        """Update a collection record with data from command line arguments."""
        if (period := self.args.full_history_retention_period) is not None:
            if period == "-":
                collection.full_history_retention_period = None
            else:
                collection.full_history_retention_period = int(period)

        if (period := self.args.metadata_only_retention_period) is not None:
            if period == "-":
                collection.metadata_only_retention_period = None
            else:
                collection.metadata_only_retention_period = int(period)


class Create(
    SingleCollectionCommand,
    RequiredInputDataCommand,
    CollectionInput,
    group="collection",
):
    """Create a new collection."""

    @override
    def update_collection(
        self, collection: CollectionDataNew | CollectionData
    ) -> None:
        super().update_collection(collection)
        collection.data = self.input_data

    @override
    def run(self) -> None:
        collection = CollectionDataNew(
            category=self.collection_category,
            name=self.collection_name,
            data={},
        )
        self.update_collection(collection)

        with self._api_call_or_fail():
            collection_new = self.debusine.collection_create(
                self.workspace, collection
            )

        self.show(collection_new)


class Manage(
    SingleCollectionCommand,
    OptionalInputDataCommand,
    CollectionInput,
    group="collection",
):
    """Configure an existing collection."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument("--rename", type=str, help="rename collection")

    @override
    def update_collection(
        self, collection: CollectionDataNew | CollectionData
    ) -> None:
        super().update_collection(collection)
        if self.input_data is not None:
            collection.data = self.input_data

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            collection = self.debusine.collection_get(
                self.workspace, self.collection_category, self.collection_name
            )
        self.update_collection(collection)

        if (name := self.args.rename) is not None:
            collection.name = name

        with self._api_call_or_fail():
            collection = self.debusine.collection_update(
                self.workspace, collection
            )

        self.show(collection)


class Delete(SingleCollectionCommand, WorkspaceCommand, group="collection"):
    """Delete a collection."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="force deleting non-empty collections",
        )

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            collection = self.debusine.collection_get(
                self.workspace,
                self.collection_category,
                self.collection_name,
                with_stats=not self.args.force,
            )

        if not self.args.force:
            assert collection.stats is not None
            items: list[str] = []
            for s in collection.stats:
                state = "active" if s["active"] else "deleted"
                items.append(
                    f"{state} {s["category"]} {s["type"]}: {s["count"]}"
                )
            if items:
                self._fail(
                    f"Collection is not empty ({", ".join(items)}):"
                    " use --force to delete it."
                )

        with self._api_call_or_fail():
            self.debusine.collection_delete(self.workspace, collection)

        self.feedback("Collection deleted.")


class LookupCommand(WorkspaceCommand, abc.ABC):
    """Base class for Lookup commands."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "--workflow",
            type=int,
            metavar="ID",
            help=(
                "Execute the lookup within workflow ID. Specify this to query "
                "the internal workflow's internal collection."
            ),
        )
        parser.add_argument(
            "--default-category",
            type=CollectionCategory,
            metavar="CATEGORY",
            help=(
                "Limit search to collections of category CATEGORY, "
                "if not specified in the lookup."
            ),
        )
        parser.add_argument(
            "--expect-type",
            type=LookupChildType,
            default=LookupChildType.ANY,
            metavar="TYPE",
            help="Limit results to items of TYPE. (Default: ANY)",
        )

    def _list_yaml(
        self, result: LookupMultipleResponse[LookupSingleResponse]
    ) -> None:
        self._print_yaml(
            result.model_dump(exclude_unset=True, exclude_none=True)
        )

    def _list_rich(
        self, results: LookupMultipleResponse[LookupSingleResponse]
    ) -> None:
        table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
        table.add_column("Type")
        table.add_column("Artifact")
        table.add_column("Collection")
        table.add_column("Parent Collection")
        table.add_column("Collection Item")
        for result in results:
            artifact = ""
            if result.artifact:
                url = self.debusine.webui_url(result.get_link_url("webui_self"))
                artifact = f"[link={url}]{result.artifact}[/]"

            collection = ""
            if result.collection:
                url = self.debusine.webui_url(result.get_link_url("webui_self"))
                collection = f"[link={url}]{result.collection}[/]"

            parent_collection = ""
            if result.parent_collection_lookup:
                lookup = result.parent_collection_lookup
                url = self.debusine.webui_url(
                    result.get_link_url("webui_parent_collection")
                )
                parent_collection = f"[link={url}]{lookup}[/]"

            table.add_row(
                result.result_type.name,
                artifact,
                collection,
                parent_collection,
                (
                    rich.json.JSON.from_data(
                        result.collection_item.model_dump()
                    )
                    if result.collection_item
                    else ""
                ),
            )
        rich.print(table)


class LookupSingleCommand(
    LookupCommand, name="lookup-single", group="collection"
):
    """Perform a lookup with a single result."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "lookup",
            type=str,
            metavar="lookup",
            help="Lookup string",
        )

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            try:
                result = self.debusine.lookup_single(
                    self.args.lookup,
                    self.args.workflow,
                    self.workspace,
                    self.args.expect_type,
                    self.args.default_category,
                )
            except DebusineError as err:
                output = {"result": "failure", "error": err.asdict()}
                self._print_yaml(output, file=sys.stderr)
                raise SystemExit(3)
        if self.args.yaml:
            self._print_yaml(
                result.model_dump(exclude_unset=True, exclude_none=True)
            )
        else:
            self._list_rich(LookupMultipleResponse([result]))


class LookupMultipleCommand(
    LookupCommand, InputDataCommand, name="lookup-multiple", group="collection"
):
    """Perform a lookup with multiple results."""

    lookup: LookupMultiple

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "lookup",
            type=str,
            metavar="lookup",
            nargs="*",
            help="Lookup string",
        )
        parser.add_argument(
            "--data",
            type=str,
            default="-",
            help=(
                "File path (or - for stdin) to read the lookups from. "
                "YAML format with a single key: lookups."
            ),
        )

    def __init__(
        self, parser: argparse.ArgumentParser, args: argparse.Namespace
    ) -> None:
        """Read lookups from input."""
        super().__init__(parser, args)
        if self.args.lookup != []:
            self.lookup = self.args.lookup
        else:
            data = self.read_input_data(self.args.data)
            if "lookups" not in data:
                sys.stderr.write("Key 'lookups' not found in YAML input.\n")
                raise SystemExit(2)
            if not isinstance(data["lookups"], list):
                sys.stderr.write("Key 'lookups' must be a list of lookups.\n")
                raise SystemExit(2)
            self.lookup = cast(LookupMultiple, data["lookups"])

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            try:
                result = self.debusine.lookup_multiple(
                    self.lookup,
                    self.args.workflow,
                    self.workspace,
                    self.args.expect_type,
                    self.args.default_category,
                )
            except DebusineError as err:
                output = {"result": "failure", "error": err.asdict()}
                self._print_yaml(output, file=sys.stderr)
                raise SystemExit(3)
        if self.args.yaml:
            self._list_yaml(result)
        else:
            self._list_rich(result)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Debusine wait tasks.

These represent steps in workflows where debusine needs to wait until
something else happens.
"""

from debusine.server.tasks.wait.base import BaseWaitTask  # isort: split

# Sub-tasks need to be imported in order to be available to DBTask
# (e.g. for DBTask.is_valid_task_name). They are registered via
# DBTask.__init_subclass__.
from debusine.server.tasks.wait.confirm import Confirm
from debusine.server.tasks.wait.delay import Delay
from debusine.server.tasks.wait.external_debsign import ExternalDebsign

__all__ = [
    "BaseWaitTask",
    "Confirm",
    "Delay",
    "ExternalDebsign",
]



# Generated by Django 5.2.7 on 2026-04-17 22:42

import django.db.models.deletion
import pgtrigger.compiler
import pgtrigger.migrations
from django.db import migrations, models

import debusine.artifacts.models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0052_collectionrelation_cascade'),
    ]

    operations = [
        pgtrigger.migrations.RemoveTrigger(
            model_name='token',
            name='db_token_required_type_parameters',
        ),
        migrations.RemoveConstraint(
            model_name='asset',
            name='db_asset_workspace_not_null',
        ),
        migrations.RemoveConstraint(
            model_name='token',
            name='db_token_consistent_type',
        ),
        migrations.AddField(
            model_name='token',
            name='work_request',
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name='work_request_token',
                to='db.workrequest',
            ),
        ),
        migrations.AlterField(
            model_name='collection',
            name='child_artifacts',
            field=models.ManyToManyField(
                related_name='parent_collections',
                through='db.CollectionItem',
                through_fields=('parent_collection', 'artifact'),
                to='db.artifact',
            ),
        ),
        migrations.AlterField(
            model_name='collection',
            name='child_collections',
            field=models.ManyToManyField(
                related_name='parent_collections',
                through='db.CollectionItem',
                through_fields=('parent_collection', 'collection'),
                to='db.collection',
            ),
        ),
        migrations.AlterField(
            model_name='token',
            name='token_type',
            field=models.CharField(
                choices=[
                    ('worker-activation', 'Worker activation'),
                    ('worker', 'Worker'),
                    ('work-request', 'Work request'),
                    ('user', 'User'),
                ]
            ),
        ),
        migrations.AlterField(
            model_name='workspace',
            name='inherits',
            field=models.ManyToManyField(
                related_name='inherited_by',
                through='db.WorkspaceChain',
                through_fields=('child', 'parent'),
                to='db.workspace',
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='token',
            trigger=pgtrigger.compiler.Trigger(
                name='db_token_required_type_parameters',
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    constraint='CONSTRAINT',
                    func=(
                        "SELECT * FROM db_token INTO NEW WHERE id = "
                        "db_token.id; IF NEW.token_type = 'worker-activation' "
                        "AND NEW.activating_worker_id IS NULL THEN RAISE "
                        "EXCEPTION 'Worker activation token % has no "
                        "activating worker', NEW.id; ELSIF NEW.token_type = "
                        "'worker' AND NEW.worker_id IS NULL THEN RAISE "
                        "EXCEPTION 'Worker token % has no worker', NEW.id; "
                        "ELSIF NEW.token_type = 'work-request' AND "
                        "NEW.work_request_id IS NULL THEN RAISE EXCEPTION "
                        "'Work request token % has no work request', NEW.id; "
                        "ELSIF NEW.token_type = 'user' AND NEW.user_id IS "
                        "NULL THEN RAISE EXCEPTION 'User token % has no user', "
                        "NEW.id; ELSE RETURN NULL; END IF;"
                    ),
                    hash='253d5833247a1e947d8b4e43a82d7f744ac2ff89',
                    operation='INSERT OR UPDATE',
                    pgid='pgtrigger_db_token_required_type_parameters_a5603',
                    table='db_token',
                    timing='DEFERRABLE INITIALLY DEFERRED',
                    when='AFTER',
                ),
            ),
        ),
        migrations.AddConstraint(
            model_name='asset',
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(
                        (
                            'category__in',
                            {
                                debusine.assets.models.AssetCategory[
                                    'APT_AUTHENTICATION'
                                ],
                                debusine.assets.models.AssetCategory[
                                    'SIGNING_KEY'
                                ],
                            },
                        ),
                        _negated=True,
                    ),
                    ('workspace__isnull', False),
                    _connector='OR',
                ),
                name='db_asset_workspace_not_null',
            ),
        ),
        migrations.AddConstraint(
            model_name='token',
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(
                        ('_worker_id__isnull', True),
                        ('token_type', 'worker-activation'),
                        ('user__isnull', True),
                        ('work_request_id__isnull', True),
                    ),
                    models.Q(
                        ('_activating_worker_id__isnull', True),
                        ('token_type', 'worker'),
                        ('user__isnull', True),
                        ('work_request_id__isnull', True),
                    ),
                    models.Q(
                        ('_activating_worker_id__isnull', True),
                        ('_worker_id__isnull', True),
                        ('token_type', 'work-request'),
                        ('user__isnull', True),
                    ),
                    models.Q(
                        ('_activating_worker_id__isnull', True),
                        ('_worker_id__isnull', True),
                        ('token_type', 'user'),
                        ('work_request_id__isnull', True),
                    ),
                    _connector='OR',
                ),
                name='db_token_consistent_type',
            ),
        ),
    ]
